import json
from typing import Union

from django.conf import settings
from django.http import HttpResponseBadRequest, JsonResponse
from django.utils.decorators import method_decorator
from django.views.decorators.csrf import csrf_exempt
from django.views.generic import View

from django_aws_api_gateway_websockets.models import ApiGateway, WebSocketSession


@method_decorator(csrf_exempt, name="dispatch")
class WebSocketView(View):
    """The base WebSocket View for handling messages sent from the client via AWS API Gateway.
    Expects a URL slug parameter called route
    """

    debug = False
    debug_log = None
    route_selection_key = "action"
    model = None
    body = {}
    aws_api_gateway_id = None  # Set to None to allow all
    api_gateway = None
    websocket_session = None
    required_headers = [
        "Host",
        "X-Forwarded-For",
        "X-Forwarded-Proto",
        "Content-Length",
        "Connectionid",
        "User-Agent",
        "X-Amzn-Apigateway-Api-Id",
    ]
    additional_required_headers = [
        "Connection",
        "X-Amzn-Trace-Id",
        "X-Forwarded-Port",
        "X-Real-Ip",
    ]
    required_connection_headers = [
        "Cookie",
        "Origin",
        "Sec-Websocket-Extensions",
        "Sec-Websocket-Key",
        "Sec-Websocket-Version",
    ]
    expected_useragent_prefix = "AmazonAPIGateway_"

    def __init__(self, **kwargs):
        self.api_gateway = kwargs.get("api_gateway", False)
        self.websocket_session = kwargs.get("websocket_session", False)
        self.debug = kwargs.get("debug", False)
        self.debug_log = []
        super().__init__(**kwargs)

    def _debug(self, msg: str):
        if self.debug:
            self.debug_log.append(msg)

    def setup(self, request, *args, **kwargs):
        """Converts the request.body string back into a dictionary and assign to the objects body property for ease"""
        super().setup(request, *args, **kwargs)
        self._debug("Within setup")
        self.body = json.loads(request.body) if request.body else {}
        self._debug("Setup completed")

    def _return_bad_request(self, msg):
        """Common method for logging and returning the HTTP400 response"""
        return HttpResponseBadRequest(msg)

    def route_selection_key_missing(
        self, request, *args, **kwargs
    ) -> HttpResponseBadRequest:
        """Method for handling missing route_selection_key"""
        msg = f"route_select_key {self.route_selection_key} missing from request body."
        self._debug(msg)
        return self._return_bad_request(msg)

    def missing_headers(self, request, *args, **kwargs) -> HttpResponseBadRequest:
        """Method for handling missing headers"""
        msg = (
            f"Some of the required headers are missing; Expected {self.required_headers}, "
            f"Received {request.headers.keys()}"
        )
        self._debug(msg)
        return self._return_bad_request("Some of the required headers are missing")

    def invalid_useragent(self, request, *args, **kwargs) -> HttpResponseBadRequest:
        """Method for handling unexpected useragents"""
        msg = (
            f"Unexpected Useragent; Expected {self.expected_useragent_prefix}{self.aws_api_gateway_id}, "
            f"Received {request.headers['User-Agent']}"
        )
        self._debug(msg)
        return self._return_bad_request("Unexpected Useragent")

    def _expected_headers(self, request, *args, **kwargs) -> bool:
        """Ensure that all required headers exist within the request header"""
        request_headers = request.headers.keys()
        res = all(h in request_headers for h in self.required_headers)

        if res and self.additional_required_headers:
            res = all(h in request_headers for h in self.additional_required_headers)

        self._debug(f"_expected_headers() returned {res}")
        return res

    def _allowed_apigateway(self, request, *args, **kwargs) -> bool:
        """Ensure the AWS API Gateway is the expected one (if set against the class) otherwise use the DB check"""
        res = self._check_platform_registered_api_gateways(request)
        if self.aws_api_gateway_id:
            res = self.aws_api_gateway_id == request.headers["X-Amzn-Apigateway-Api-Id"]
            self._debug(f"_expected_apigateway_id() returned {res}")
        self._debug(f"_allowed_apigateway() returned {res}")
        return res

    def _check_platform_registered_api_gateways(self, request) -> bool:
        """Checks to ensure that the API Gateway calling the view is one that the user has registered"""
        self.api_gateway = ApiGateway.objects.filter(
            api_id=request.headers["X-Amzn-Apigateway-Api-Id"]
        ).first()
        self._debug(
            f"_check_platform_registered_api_gateways() returned {bool(self.api_gateway)}"
        )
        return bool(self.api_gateway)

    def _expected_useragent(self, request, *args, **kwargs) -> bool:
        """Validated that the useragent is the expected one for all calls except the connect method

        For the connect method the useragent should be API Gateway itself and NOT the client's forwarded useragent
        """
        if self.aws_api_gateway_id:
            res = (
                not request.headers["User-Agent"]
                == f"{self.expected_useragent_prefix}{self.aws_api_gateway_id}"
            )
        else:
            res = self.expected_useragent_prefix in request.headers["User-Agent"]
        self._debug(f"_expected_useragent() returned {res}")

        return res

    @staticmethod
    def _check_allowed_hosts(request) -> bool:
        """Check that the host is within the allowed hosts"""
        if (
            settings.ALLOWED_HOSTS
            and request.headers["Host"] not in settings.ALLOWED_HOSTS
        ):
            return False
        return True

    @staticmethod
    def _check_host_is_in_origin(request) -> bool:
        """Check that the value of the Host header is within the Origin header. Origin will have the protocol as well"""
        if request.headers["Host"] not in request.headers["Origin"]:
            return False
        return True

    def _expected_connection_headers(self, request, *args, **kwargs) -> bool:
        """Run additional checks for the connection route for security"""
        request_headers = request.headers.keys()

        return all(h in request_headers for h in self.required_connection_headers)

    def _add_user_to_request(self, request):
        """Fetch the user from the model and append it back into the request variable"""
        wss = WebSocketSession.objects.get(
            connection_id=request.headers["Connectionid"]
        )
        if wss.user:
            request.user = wss.user
        wss.request_count += 1
        wss.save()

    def _get_channel_name(self, request) -> str:
        """Returns the name of the channel to use

        The "channel" can be optionally be set as a querystring parameter during the connection. If it is not but the
        api_gateway that was selected has a default_channel_name set then that will be used instead. Otherwise an empty
        string is returned
        """
        channel_name = request.GET.get("channel", "")
        if (
            not channel_name
            and self.api_gateway
            and self.api_gateway.default_channel_name
        ):
            channel_name = self.api_gateway.default_channel_name

        return channel_name

    def _load_session(self, request):
        self.websocket_session = WebSocketSession.objects.get(
            connection_id=request.headers["Connectionid"]
        )

    def dispatch(self, request, *args, **kwargs):
        """Determine the correct method to call. The method will map to the route_selection_key or default.

        Checks for the expected headers.
        Tries to dispatch to the right method; if a method doesn't exist defer to the default handler.
        If the Route Selection Key is missing defer to the route selection error handler.
        If the request method isn't on the approved list then defer to the normal error handler .
        """
        if self._expected_headers(request) and self._allowed_apigateway(request):
            if request.method.lower() in self.http_method_names:
                if "connect" == self.kwargs.get("route"):
                    handler = self.connect
                elif "disconnect" == self.kwargs.get("route"):
                    if not self._expected_useragent(request, *args, **kwargs):
                        handler = self.invalid_useragent
                    else:
                        handler = self.disconnect
                        self._add_user_to_request(request)
                elif self.route_selection_key in self.body:
                    self._load_session(request)

                    handler = getattr(
                        self, self.body[self.route_selection_key], self.default
                    )
                    if not self._expected_useragent(request, *args, **kwargs):
                        handler = self.invalid_useragent
                    else:
                        self._add_user_to_request(request)
                else:
                    handler = self.route_selection_key_missing
            else:
                handler = self.http_method_not_allowed
        else:
            handler = self.missing_headers
        res = handler(request, *args, **kwargs)
        return res if res else JsonResponse({})

    def connect(
        self, request, *args, **kwargs
    ) -> Union[JsonResponse, HttpResponseBadRequest]:
        """Handle the connection route in a standard way that ensures the User to Connectionid mapping persists"""
        if not self._expected_connection_headers(request, *args, **kwargs):
            request_headers = request.headers.keys()
            missing_headers = [
                h for h in self.required_connection_headers if h not in request_headers
            ]
            self._debug(
                f"Missing headers; Expected {self.required_connection_headers}, Received {request.headers}"
            )
            msg = f"Missing {len(missing_headers)} headers"
            return self._return_bad_request(msg)

        if not self._check_allowed_hosts(request):
            msg = f"Host is not in AllowedHosts {settings.ALLOWED_HOSTS}"
            self._debug(msg)
            return self._return_bad_request(msg)

        if not self._check_host_is_in_origin(request):
            self._debug("Host is not in Origin")
            return self._return_bad_request("Host is not in Origin")

        res, msg = self._additional_connection_checks(request, *args, **kwargs)
        if not res:
            return self._return_bad_request(msg)

        WebSocketSession.objects.create(
            connection_id=request.headers["Connectionid"],
            channel_name=self._get_channel_name(request),
            user=request.user if request.user.is_authenticated else None,
            api_gateway=self.api_gateway,
        )

        return JsonResponse({})

    def _additional_connection_checks(self, request, *args, **kwargs) -> (bool, str):
        """Could add in additional steps for certificates, APIGateway Authorizers etc"""
        return True, ""

    def disconnect(self, request, *args, **kwargs):
        """Using connectionId update websocket table to show as disconnected"""
        wss = WebSocketSession.objects.get(
            connection_id=request.headers["Connectionid"]
        )
        wss.connected = False
        wss.save()

    def default(self, request, *args, **kwargs) -> JsonResponse:
        """Overload this method if you want to have a default message handler"""
        raise NotImplementedError("This logic needs to be defined within the subclass")



import json
from unittest.mock import MagicMock, patch

from django.core.exceptions import ObjectDoesNotExist
from django.http import HttpResponseBadRequest, HttpResponseNotAllowed, JsonResponse
from django.test import RequestFactory, SimpleTestCase, override_settings

from django_aws_api_gateway_websockets import views
from django_aws_api_gateway_websockets.models import ApiGateway


class WebSocketViewSimpleTestCase(SimpleTestCase):
    """For tests that do not require a DB"""

    def setUp(self) -> None:
        self.factory = RequestFactory()

    def test_default_required_headers(self):
        self.assertEqual(
            [
                "Host",
                "X-Forwarded-For",
                "X-Forwarded-Proto",
                "Content-Length",
                "Connectionid",
                "User-Agent",
                "X-Amzn-Apigateway-Api-Id",
            ],
            views.WebSocketView.required_headers,
        )

    def test_additional_required_headers(self):
        self.assertEqual(
            [
                "Connection",
                "X-Amzn-Trace-Id",
                "X-Forwarded-Port",
                "X-Real-Ip",
            ],
            views.WebSocketView.additional_required_headers,
        )

    def test_default_required_connection_headers(self):
        self.assertEqual(
            [
                "Cookie",
                "Origin",
                "Sec-Websocket-Extensions",
                "Sec-Websocket-Key",
                "Sec-Websocket-Version",
            ],
            views.WebSocketView.required_connection_headers,
        )

    def test_default_expected_useragent_prefix(self):
        self.assertEqual(
            "AmazonAPIGateway_", views.WebSocketView.expected_useragent_prefix
        )

    def test_default_aws_api_gateway_id(self):
        self.assertIsNone(views.WebSocketView.aws_api_gateway_id)

    def test__debug_when_debug_is_false(self):
        view = views.WebSocketView()
        view.debug = False
        view._debug("A message")
        self.assertEqual([], view.debug_log)

    def test__debug_when_debug_is_true(self):
        """Should append to the debug log"""
        view = views.WebSocketView()
        view.debug = True
        view._debug("A message")
        self.assertEqual(["A message"], view.debug_log)
        view._debug("Another entry")
        self.assertEqual(["A message", "Another entry"], view.debug_log)

    def test_default_route_selection_key(self):
        self.assertEqual("action", views.WebSocketView.route_selection_key)

    def test_setup(self):
        """Ensure the setup is working as expected by turning on debug and looking for the records"""
        post_data = {"action": "default", "key": "value"}
        request = self.factory.post(
            "", data=json.dumps(post_data), content_type="application/json"
        )

        view = views.WebSocketView()
        view.debug = True
        view.setup(request=request)

        self.assertEqual(["Within setup", "Setup completed"], view.debug_log)

    def test__return_bad_request(self):
        """Should log a warning message and return a HttpBadRequest object"""
        msg = "This is the reason for the bad request"
        obj = views.WebSocketView()
        res = obj._return_bad_request(msg)

        self.assertIsInstance(res, HttpResponseBadRequest)
        self.assertEqual(400, res.status_code)

    def test_route_selection_key_missing(self):
        """Ensure the expected response is generated when the route selection key is missing"""
        obj = views.WebSocketView()
        res = obj.route_selection_key_missing(None)

        self.assertIsInstance(res, HttpResponseBadRequest)
        self.assertEqual(400, res.status_code)
        self.assertEqual(
            b"route_select_key action missing from request body.", res.content
        )

    def test_missing_headers(self):
        """Ensure the expected response is generated when there are missing headers"""
        post_data = {"action": "default", "key": "value"}
        request = self.factory.post(
            "", data=json.dumps(post_data), content_type="application/json"
        )

        obj = views.WebSocketView()
        res = obj.missing_headers(request=request)

        self.assertIsInstance(res, HttpResponseBadRequest)
        self.assertEqual(400, res.status_code)
        self.assertEqual(b"Some of the required headers are missing", res.content)

    def test_invalid_useragent(self):
        """Ensure the expected response is generated when the useragent is invalid"""
        post_data = {"action": "default", "key": "value"}
        request = self.factory.post(
            "",
            data=json.dumps(post_data),
            content_type="application/json",
            HTTP_USER_AGENT="Mozilla/5.0",
        )

        obj = views.WebSocketView()
        res = obj.invalid_useragent(request=request)

        self.assertIsInstance(res, HttpResponseBadRequest)
        self.assertEqual(400, res.status_code)
        self.assertEqual(b"Unexpected Useragent", res.content)

    def test__expected_headers(self):
        """Ensure the methods returns True when all headers are present otherwise should return False"""
        post_data = {"action": "default", "key": "value"}

        test_headers = [
            {
                "required": ["My-Expected-Header", "Another-Header"],
                "additional_required": ["X-Amzn-Trace-Id", "Connection"],
                "supplied": {
                    "HTTP_My_Expected_Header": "some value",
                    "HTTP_X_Amzn_Trace_Id": "err-gf234",
                    "HTTP_Connection": "SFfgGFG",
                },
                "expected_result": False,
            },
            {
                "required": ["My-Expected-Header", "Another-Header"],
                "additional_required": ["X-Amzn-Trace-Id", "Connection"],
                "supplied": {
                    "HTTP_My_Expected_Header": "some value",
                    "HTTP_Another_Header": "some value",
                    "HTTP_X_Amzn_Trace_Id": "err-gf234",
                    "HTTP_Connection": "SFfgGFG",
                },
                "expected_result": True,
            },
        ]
        for headers in test_headers:
            with self.subTest(headers=headers):
                obj = views.WebSocketView()
                obj.required_headers = headers["required"]
                obj.additional_required_headers = headers["additional_required"]
                request = self.factory.post(
                    "",
                    data=json.dumps(post_data),
                    content_type="application/json",
                    HTTP_USER_AGENT="Mozilla/5.0",
                    **headers["supplied"],
                )

                self.assertEqual(
                    headers["expected_result"], obj._expected_headers(request=request)
                )

    @patch("django_aws_api_gateway_websockets.views.ApiGateway")
    def test_apigateway_check_when_db_record_matches_header_and_class_attribute_is_not_set_return_true(
        self, MockApiGateway
    ):
        """
        When the class does not have an aws_api_gateway_id attribute set
        and the X_Amzn_Apigateway_Api_Id header DOES MATCH a database record

        The method should return True I.E accept the request as it's from an expected API Gateway endpoint

        :param MockApiGateway:
        :return:
        """
        allowed_api_id = "ABC123AB"
        api_gateway = MagicMock(api_id=allowed_api_id)
        MockApiGateway.objects.filter.return_value.first.return_value = api_gateway

        post_data = {"action": "default", "key": "value"}
        request = self.factory.post(
            "",
            data=json.dumps(post_data),
            content_type="application/json",
            HTTP_USER_AGENT="Mozilla/5.0",
            HTTP_X_Amzn_Apigateway_Api_Id=allowed_api_id,
        )

        obj = views.WebSocketView()
        obj.aws_api_gateway_id = None
        obj._allowed_apigateway(request)

        MockApiGateway.objects.filter.assert_called_with(
            api_id=request.headers["X-Amzn-Apigateway-Api-Id"]
        )
        MockApiGateway.objects.filter.return_value.first.assert_called_with()
        self.assertTrue(obj._allowed_apigateway(request=request))

    @patch("django_aws_api_gateway_websockets.views.ApiGateway")
    def test_apigateway_check_when_db_record_does_not_match_header_and_class_attribute_is_not_set_return_false(
        self, MockApiGateway
    ):
        """
        When the class does not have an aws_api_gateway_id attribute set
        and the X_Amzn_Apigateway_Api_Id header does not match ANY database record

        The method should return False I.E do not accept the request
        """
        MockApiGateway.objects.filter.return_value.first.return_value = None

        post_data = {"action": "default", "key": "value"}
        request = self.factory.post(
            "",
            data=json.dumps(post_data),
            content_type="application/json",
            HTTP_USER_AGENT="Mozilla/5.0",
            HTTP_X_Amzn_Apigateway_Api_Id="some-other-api-gatewway",
        )

        obj = views.WebSocketView()
        obj.aws_api_gateway_id = None
        obj._allowed_apigateway(request)

        MockApiGateway.objects.filter.assert_called_with(
            api_id=request.headers["X-Amzn-Apigateway-Api-Id"]
        )
        MockApiGateway.objects.filter.return_value.first.assert_called_with()

        self.assertFalse(obj._allowed_apigateway(request=request))

    @patch("django_aws_api_gateway_websockets.views.ApiGateway")
    def test_apigateway_check_when_db_record_matches_header_and_the_class_attribute_return_true(
        self, MockApiGateway
    ):
        """
        When the class has an aws_api_gateway_id attribute set and it matches BOTH the header and the DB record

        The method should return True I.E accept the request as it's from an expected API Gateway endpoint

        :param MockApiGateway:
        :return:
        """
        allowed_api_id = "ABC123AB"
        api_gateway = MagicMock(api_id=allowed_api_id)
        MockApiGateway.objects.filter.return_value.first.return_value = api_gateway

        post_data = {"action": "default", "key": "value"}
        request = self.factory.post(
            "",
            data=json.dumps(post_data),
            content_type="application/json",
            HTTP_USER_AGENT="Mozilla/5.0",
            HTTP_X_Amzn_Apigateway_Api_Id=allowed_api_id,
        )

        obj = views.WebSocketView()
        obj.aws_api_gateway_id = allowed_api_id
        obj._allowed_apigateway(request)

        MockApiGateway.objects.filter.assert_called_with(
            api_id=request.headers["X-Amzn-Apigateway-Api-Id"]
        )
        MockApiGateway.objects.filter.return_value.first.assert_called_with()
        self.assertTrue(obj._allowed_apigateway(request=request))

    @patch("django_aws_api_gateway_websockets.views.ApiGateway")
    def test_apigateway_no_db_match_found_and_class_attribute_not_matched_return_false(
        self, MockApiGateway
    ):
        """
        When the class has an aws_api_gateway_id attribute set but it does not match the header
        And the DB check also does not match

        Return False
        """
        MockApiGateway.objects.filter.return_value.first.return_value = None

        post_data = {"action": "default", "key": "value"}
        request = self.factory.post(
            "",
            data=json.dumps(post_data),
            content_type="application/json",
            HTTP_USER_AGENT="Mozilla/5.0",
            HTTP_X_Amzn_Apigateway_Api_Id="some-other-api-gatewway",
        )

        obj = views.WebSocketView()
        obj.aws_api_gateway_id = "some-other-id"
        obj._allowed_apigateway(request)

        MockApiGateway.objects.filter.assert_called_with(
            api_id=request.headers["X-Amzn-Apigateway-Api-Id"]
        )
        MockApiGateway.objects.filter.return_value.first.assert_called_with()

        self.assertFalse(obj._allowed_apigateway(request=request))

    @patch("django_aws_api_gateway_websockets.views.ApiGateway")
    def test_apigateway_with_db_match_but_class_attribute_not_matched_return_false(
        self, MockApiGateway
    ):
        """
        When the class has an aws_api_gateway_id attribute set but it does not match the header

        The method should return False even if the DB record matches
        """
        allowed_api_id = "ABC123AB"
        api_gateway = MagicMock(api_id=allowed_api_id)
        MockApiGateway.objects.filter.return_value.first.return_value = api_gateway

        post_data = {"action": "default", "key": "value"}
        request = self.factory.post(
            "",
            data=json.dumps(post_data),
            content_type="application/json",
            HTTP_USER_AGENT="Mozilla/5.0",
            HTTP_X_Amzn_Apigateway_Api_Id=allowed_api_id,
        )

        obj = views.WebSocketView()
        obj.aws_api_gateway_id = "some-other-id"
        obj._allowed_apigateway(request)

        MockApiGateway.objects.filter.assert_called_with(
            api_id=request.headers["X-Amzn-Apigateway-Api-Id"]
        )
        MockApiGateway.objects.filter.return_value.first.assert_called_with()

        self.assertFalse(obj._allowed_apigateway(request=request))

    def test__expected_useragent_returns_true_when_aws_api_key_is_set_and_is_in_the_user_agent(
        self,
    ):
        """_expected_useragent should return False if the AWS API Gateway is making the request

        That is determined by the useragent matching a particular string of "AmazonAPIGateway_" followed the API ID
        of the API Gateway. Only the "connect" call should use that user agent, all other connections should either
        use the REAL client useragent or nothing. In this test we are testing as if the API Gateway user agent was used
        """
        allowed_api_id = "ABC123AB"

        post_data = {"action": "default", "key": "value"}
        request = self.factory.post(
            "",
            data=json.dumps(post_data),
            content_type="application/json",
            HTTP_USER_AGENT=f"AmazonAPIGateway_{allowed_api_id}",
        )

        obj = views.WebSocketView()
        obj.aws_api_gateway_id = allowed_api_id

        self.assertFalse(obj._expected_useragent(request=request))

    def test__expected_useragent_returns_true_when_useragent_is_not_the_exepcted_api_gateways_useragent(
        self,
    ):
        """
        If the aws_api_gateway_id is set and the useragent is NOT the useragent we expect to receive from the API Gateway
        I.E AmazonAPIGateway_ then a API Gateway ID, then the method should return True
        """
        allowed_api_id = "ABC123AB"

        post_data = {"action": "default", "key": "value"}
        request = self.factory.post(
            "",
            data=json.dumps(post_data),
            content_type="application/json",
            HTTP_USER_AGENT="AmazonAPIGateway_ANOTHER_ID",
        )

        obj = views.WebSocketView()
        obj.aws_api_gateway_id = allowed_api_id

        self.assertTrue(obj._expected_useragent(request=request))

    def test__expected_useragent_returns_false_when_useragent_is_not_the_exepcted_api_gateways_useragent(
        self,
    ):
        """
        If the aws_api_gateway_id is set and the user agent is the useragent we expect to receive from the API Gateway
        I.E where the useragent is AmazonAPIGateway_ then a API Gateway ID, then the method should return False
        """
        allowed_api_id = "ABC123AB"

        post_data = {"action": "default", "key": "value"}
        request = self.factory.post(
            "",
            data=json.dumps(post_data),
            content_type="application/json",
            HTTP_USER_AGENT=f"AmazonAPIGateway_{allowed_api_id}",
        )

        obj = views.WebSocketView()
        obj.aws_api_gateway_id = allowed_api_id

        self.assertFalse(obj._expected_useragent(request=request))

    def test__expected_useragent_with_no_allowed_api_id_but_an_apigateway_made_the_call__should_return_true(
        self,
    ):
        """
        If the aws_api_gateway_id is set and the user agent is the useragent we expect to receive from the API Gateway
        I.E where the useragent is AmazonAPIGateway_ then a API Gateway ID, then the method should return False
        """
        allowed_api_id = ""

        post_data = {"action": "default", "key": "value"}
        request = self.factory.post(
            "",
            data=json.dumps(post_data),
            content_type="application/json",
            HTTP_USER_AGENT=f"AmazonAPIGateway_SomeOtherId",
        )

        obj = views.WebSocketView()
        obj.aws_api_gateway_id = allowed_api_id

        self.assertTrue(obj._expected_useragent(request=request))

    def test__expected_useragent_with_no_allowed_api_id_and_an_apigateway_didnt_make_the_call__should_return_false(
        self,
    ):
        """
        If the aws_api_gateway_id is not set and the user agent is a normal user then False should be returned
        """
        allowed_api_id = ""

        post_data = {"action": "default", "key": "value"}
        request = self.factory.post(
            "",
            data=json.dumps(post_data),
            content_type="application/json",
            HTTP_USER_AGENT=f"Mozzila/5.0",
        )

        obj = views.WebSocketView()
        obj.aws_api_gateway_id = allowed_api_id

        self.assertFalse(obj._expected_useragent(request=request))

    @override_settings(ALLOWED_HOSTS=["www.example.com"])
    def test__check_allowed_hosts__when_host_is_allowed__return_true(self):
        """When the host of the request is within the allowed hosts of settings.py True should be returned"""
        post_data = {"action": "default", "key": "value"}
        request = self.factory.post(
            "",
            data=json.dumps(post_data),
            content_type="application/json",
            HTTP_USER_AGENT=f"Mozzila/5.0",
            HTTP_HOST="www.example.com",
        )

        obj = views.WebSocketView()

        self.assertTrue(obj._check_allowed_hosts(request=request))

    @override_settings(ALLOWED_HOSTS=["www.example.com"])
    def test__check_allowed_hosts__when_host_is_not_allowed__return_false(self):
        """When the host of the request is within the allowed hosts of settings.py False should be returned"""
        post_data = {"action": "default", "key": "value"}
        request = self.factory.post(
            "",
            data=json.dumps(post_data),
            content_type="application/json",
            HTTP_USER_AGENT=f"Mozzila/5.0",
            HTTP_HOST="www.anotherdomain.com",
        )

        obj = views.WebSocketView()

        self.assertFalse(obj._check_allowed_hosts(request=request))

    @override_settings(ALLOWED_HOSTS=["www.example.com"])
    def test__check_host_is_in_origin__when_it_iss(self):
        """When the host of the request is within the origin of the request True should be returned"""
        post_data = {"action": "default", "key": "value"}
        request = self.factory.post(
            "",
            data=json.dumps(post_data),
            content_type="application/json",
            HTTP_USER_AGENT=f"Mozzila/5.0",
            HTTP_HOST="www.example.com",
            HTTP_ORIGIN="https://www.example.com",
        )

        obj = views.WebSocketView()

        self.assertTrue(obj._check_host_is_in_origin(request=request))

    @override_settings(ALLOWED_HOSTS=["www.example.com"])
    def test__check_host_is_in_origin__when_it_is_not(self):
        """When the host of the request is NOT ithin the origin of the request False should be returned"""
        post_data = {"action": "default", "key": "value"}
        request = self.factory.post(
            "",
            data=json.dumps(post_data),
            content_type="application/json",
            HTTP_USER_AGENT=f"Mozzila/5.0",
            HTTP_HOST="www.anotherdomain.com",
            HTTP_ORIGIN="http://www.example.com",
        )

        obj = views.WebSocketView()

        self.assertFalse(obj._check_host_is_in_origin(request=request))

    def test__expected_connection_headers(self):
        """Ensure the methods returns True when all headers are present otherwise should return False"""
        post_data = {"action": "default", "key": "value"}

        test_headers = [
            {
                "name": "All expected headers",
                "required_connection_headers": [
                    "Cookie",
                    "Origin",
                    "Sec-Websocket-Extensions",
                    "Sec-Websocket-Key",
                    "Sec-Websocket-Version",
                ],
                "supplied": {
                    "HTTP_Cookie": "some-cookie",
                    "HTTP_Origin": "example.com",
                    "HTTP_Sec-Websocket-Extensions": "some-value",
                    "HTTP_Sec-Websocket-Key": "some-key",
                    "HTTP_Sec-Websocket-Version": "1.2.3",
                },
                "expected_result": True,
            },
            {
                "name": "Missing a header",
                "required_connection_headers": [
                    "Cookie",
                    "Origin",
                    "Sec-Websocket-Extensions",
                    "Sec-Websocket-Key",
                    "Sec-Websocket-Version",
                ],
                "supplied": {
                    "HTTP_Cookie": "some-cookie",
                    "HTTP_Origin": "example.com",
                    "HTTP_Sec-Websocket-Extensions": "some-value",
                    "HTTP_Sec-Websocket-Key": "some-key",
                },
                "expected_result": False,
            },
        ]
        for headers in test_headers:
            with self.subTest(headers=headers):
                obj = views.WebSocketView()
                obj.required_connection_headers = headers["required_connection_headers"]
                request = self.factory.post(
                    "",
                    data=json.dumps(post_data),
                    content_type="application/json",
                    HTTP_USER_AGENT="Mozilla/5.0",
                    **headers["supplied"],
                )

                self.assertEqual(
                    headers["expected_result"],
                    obj._expected_connection_headers(request=request),
                    headers["name"],
                )

    @patch("django_aws_api_gateway_websockets.views.WebSocketSession")
    def test__add_user_to_request_raises_when_object_not_found(
        self, MockWebSocketSession
    ):
        """An exception should be raised if the object can not be found"""
        MockWebSocketSession.objects.get.side_effect = ObjectDoesNotExist(
            "no match found"
        )

        con_id = "12345"

        obj = views.WebSocketView()
        request = self.factory.post(
            "",
            data=json.dumps({}),
            content_type="application/json",
            HTTP_Connectionid=con_id,
        )

        self.assertFalse(hasattr(request, "user"))
        with self.assertRaises(ObjectDoesNotExist):
            obj._add_user_to_request(request)

            MockWebSocketSession.objects.get.assert_called_with(
                connection_id=request.headers["Connectionid"]
            )

            self.assertFalse(hasattr(request, "user"))

    @patch("django_aws_api_gateway_websockets.views.WebSocketSession")
    def test__add_user_to_request_when_session_found(self, MockWebSocketSession):
        """When a WebSocketSession is found then the user will be added to the request IF it's set against the session"""
        mocked_user = MagicMock(pk=12)
        configs = [
            {"name": "No User in session", "user": None},
            {"name": "User set in session", "user": mocked_user},
        ]

        for config in configs:
            with self.subTest(config=config):
                wss = MagicMock(user=config["user"], request_count=0)
                MockWebSocketSession.objects.get.return_value = wss
                con_id = "12345"

                obj = views.WebSocketView()
                request = self.factory.post(
                    "",
                    data=json.dumps({}),
                    content_type="application/json",
                    HTTP_Connectionid=con_id,
                )

                self.assertFalse(hasattr(request, "user"))
                obj._add_user_to_request(request)

                MockWebSocketSession.objects.get.assert_called_with(
                    connection_id=request.headers["Connectionid"]
                )

                if config["user"]:
                    self.assertTrue(hasattr(request, "user"))
                    self.assertEqual(mocked_user, request.user)
                    self.assertEqual(1, wss.request_count)
                else:
                    self.assertFalse(hasattr(request, "user"))
                    self.assertEqual(1, wss.request_count)

    def test__get_channel_name(self):
        """Ensure the method returns the channel either from the GET string of set against the API Gateway"""
        api_gateway = ApiGateway(default_channel_name="default channel")
        configs = [
            {
                "name": "channel in url only",
                "qs": "channel=my-channel",
                "channel": "my-channel",
                "db": False,
            },
            {
                "name": "channel in url and db",
                "qs": "channel=my-channel",
                "channel": "my-channel",
                "db": True,
            },
            {
                "name": "channel db only",
                "qs": "",
                "channel": "default channel",
                "db": True,
            },
            {"name": "no channel", "qs": "", "channel": "", "db": False},
        ]

        for config in configs:
            with self.subTest(config=config):
                obj = views.WebSocketView()
                if config["db"]:
                    obj.api_gateway = api_gateway

                request = self.factory.post(
                    f"?{config['qs']}",
                    data=json.dumps({}),
                    content_type="application/json",
                )

                res = obj._get_channel_name(request)

                self.assertEqual(config["channel"], res, config["name"])

    @patch("django_aws_api_gateway_websockets.views.WebSocketSession")
    def test__load_session(self, MockWebSocketSession):
        """Ensure the WebSocketSession is fetched from the DB and set against the websocket_session class attribute"""
        request = self.factory.post(
            "",
            data=json.dumps({}),
            content_type="application/json",
            HTTP_Connectionid="1234",
        )
        obj = views.WebSocketView()
        obj._load_session(request)

        MockWebSocketSession.objects.get.assert_called_with(connection_id="1234")
        self.assertEqual(
            obj.websocket_session, MockWebSocketSession.objects.get.return_value
        )

    def test__additional_connection_checks(self):
        """The _additional_connection_checks method should return True, "" for the default implementation"""
        obj = views.WebSocketView()
        result, msg = obj._additional_connection_checks(None)
        self.assertTrue(result)
        self.assertEqual("", msg)

    @patch("django_aws_api_gateway_websockets.views.WebSocketSession")
    def test_disconnect(self, MockWebSocketSession):
        """The disconnect method should load the websocket session, update a property then call the save method"""
        wss = MagicMock(connected=True)
        MockWebSocketSession.objects.get.return_value = wss

        request = self.factory.post(
            "",
            data=json.dumps({}),
            content_type="application/json",
            HTTP_Connectionid="1234",
        )
        obj = views.WebSocketView()

        self.assertTrue(wss.connected)

        obj.disconnect(request)

        MockWebSocketSession.objects.get.assert_called_with(
            connection_id=request.headers["Connectionid"]
        )
        MockWebSocketSession.objects.get.return_value.save.assert_called_with()
        self.assertFalse(MockWebSocketSession.objects.get.return_value.connected)

    def test_default(self):
        """The method handling default actions should raise a NotImplemented error by default"""
        obj = views.WebSocketView()

        with self.assertRaises(NotImplementedError) as e:
            obj.default(request=self.factory.get(""))
        self.assertEqual(
            "This logic needs to be defined within the subclass", str(e.exception)
        )

    @override_settings(ALLOWED_HOSTS=["www.example.com"])
    @patch("django_aws_api_gateway_websockets.views.WebSocketSession")
    @patch("django_aws_api_gateway_websockets.views.JsonResponse")
    def test_connect(self, MockJsonResponse, MockWebSocketSession):
        """Test the connect method when the expected header are set. Should pass through all checks"""
        user = MagicMock(is_authenticated=True)
        request = self.factory.post(
            "?channel=my-channel",
            data=json.dumps({}),
            content_type="application/json",
            HTTP_Connectionid="1234",
            HTTP_Cookie="some-cookie",
            HTTP_Sec_Websocket_Extensions="some-value",
            HTTP_Sec_Websocket_Key="some-key",
            HTTP_Sec_Websocket_Version="1.2.3",
            HTTP_HOST="www.example.com",
            HTTP_ORIGIN="https://www.example.com",
        )
        request.user = user
        obj = views.WebSocketView()

        res = obj.connect(request)
        MockWebSocketSession.objects.create.assert_called_with(
            connection_id=request.headers["Connectionid"],
            channel_name=obj._get_channel_name(request),
            user=request.user if request.user.is_authenticated else None,
            api_gateway=obj.api_gateway,
        )
        MockJsonResponse.assert_called_with({})
        self.assertEqual(MockJsonResponse.return_value, res)

    @override_settings(ALLOWED_HOSTS=["www.example.com"])
    @patch("django_aws_api_gateway_websockets.views.WebSocketSession")
    @patch("django_aws_api_gateway_websockets.views.JsonResponse")
    @patch("django_aws_api_gateway_websockets.views.HttpResponseBadRequest")
    def test_connect__missing_expected_headers(
        self, MockHttpResponseBadRequest, MockJsonResponse, MockWebSocketSession
    ):
        """Test the connect method when the expected headers are missing."""
        user = MagicMock(is_authenticated=True)
        request = self.factory.post(
            "?channel=my-channel",
            data=json.dumps({}),
            content_type="application/json",
            HTTP_Connectionid="1234",
            HTTP_Cookie="some-cookie",
            HTTP_Origin="example.com",
            HTTP_HOST="www.example.com",
            HTTP_ORIGIN="https://www.example.com",
        )
        request.user = user
        obj = views.WebSocketView()

        res = obj.connect(request)

        self.assertEqual(MockHttpResponseBadRequest.return_value, res)
        self.assertEqual(0, MockWebSocketSession.objects.create.call_count)
        self.assertEqual(0, MockJsonResponse.call_count)
        MockHttpResponseBadRequest.assert_called_with("Missing 3 headers")

    @override_settings(ALLOWED_HOSTS=["www.example.com"])
    @patch("django_aws_api_gateway_websockets.views.WebSocketSession")
    @patch("django_aws_api_gateway_websockets.views.JsonResponse")
    @patch("django_aws_api_gateway_websockets.views.HttpResponseBadRequest")
    def test_connect__with_invalid_host(
        self, MockHttpResponseBadRequest, MockJsonResponse, MockWebSocketSession
    ):
        """Test the connect method when the host is not in the allowed list are missing."""
        user = MagicMock(is_authenticated=True)
        request = self.factory.post(
            "?channel=my-channel",
            data=json.dumps({}),
            content_type="application/json",
            HTTP_Connectionid="1234",
            HTTP_Cookie="some-cookie",
            HTTP_Origin="example.com",
            HTTP_Sec_Websocket_Extensions="some-value",
            HTTP_Sec_Websocket_Key="some-key",
            HTTP_Sec_Websocket_Version="1.2.3",
            HTTP_HOST="www.spoofed.com",
            HTTP_ORIGIN="https://www.spoofed.com",
        )
        request.user = user
        obj = views.WebSocketView()

        res = obj.connect(request)

        self.assertEqual(MockHttpResponseBadRequest.return_value, res)
        self.assertEqual(0, MockWebSocketSession.objects.create.call_count)
        self.assertEqual(0, MockJsonResponse.call_count)
        MockHttpResponseBadRequest.assert_called_with(
            (f"Host is not in AllowedHosts ['www.example.com']")
        )

    @override_settings(ALLOWED_HOSTS=["www.example.com"])
    @patch("django_aws_api_gateway_websockets.views.WebSocketSession")
    @patch("django_aws_api_gateway_websockets.views.JsonResponse")
    @patch("django_aws_api_gateway_websockets.views.HttpResponseBadRequest")
    def test_connect__with_host_not_in_origin(
        self, MockHttpResponseBadRequest, MockJsonResponse, MockWebSocketSession
    ):
        """Test the connect method when the host does not exist within the origin."""
        user = MagicMock(is_authenticated=True)
        request = self.factory.post(
            "?channel=my-channel",
            data=json.dumps({}),
            content_type="application/json",
            HTTP_Connectionid="1234",
            HTTP_Cookie="some-cookie",
            HTTP_Origin="example.com",
            HTTP_Sec_Websocket_Extensions="some-value",
            HTTP_Sec_Websocket_Key="some-key",
            HTTP_Sec_Websocket_Version="1.2.3",
            HTTP_HOST="www.example.com",
            HTTP_ORIGIN="https://www.spoofed.com",
        )
        request.user = user
        obj = views.WebSocketView()

        res = obj.connect(request)

        self.assertEqual(MockHttpResponseBadRequest.return_value, res)
        self.assertEqual(0, MockWebSocketSession.objects.create.call_count)
        MockHttpResponseBadRequest.assert_called_with("Host is not in Origin")
        self.assertEqual(0, MockJsonResponse.call_count)

    @override_settings(ALLOWED_HOSTS=["www.example.com"])
    @patch("django_aws_api_gateway_websockets.views.WebSocketSession")
    @patch("django_aws_api_gateway_websockets.views.JsonResponse")
    @patch("django_aws_api_gateway_websockets.views.HttpResponseBadRequest")
    def test_connect__additional_connection_checks_returns_false(
        self, MockHttpResponseBadRequest, MockJsonResponse, MockWebSocketSession
    ):
        """Test the connect method when the expected headers are missing."""
        user = MagicMock(is_authenticated=True)
        request = self.factory.post(
            "?channel=my-channel",
            data=json.dumps({}),
            content_type="application/json",
            HTTP_Connectionid="1234",
            HTTP_Cookie="some-cookie",
            HTTP_Origin="example.com",
            HTTP_Sec_Websocket_Extensions="some-value",
            HTTP_Sec_Websocket_Key="some-key",
            HTTP_Sec_Websocket_Version="1.2.3",
            HTTP_HOST="www.example.com",
            HTTP_ORIGIN="https://www.example.com",
        )
        request.user = user

        class SubClassedView(views.WebSocketView):
            def _additional_connection_checks(
                self, request, *args, **kwargs
            ) -> (bool, str):
                """Could add in additional steps for certificates, APIGateway Authorizers etc"""
                return False, "Some error message raised in subclass"

        obj = SubClassedView()

        res = obj.connect(request)

        self.assertEqual(MockHttpResponseBadRequest.return_value, res)
        self.assertEqual(0, MockWebSocketSession.objects.create.call_count)
        MockHttpResponseBadRequest.assert_called_with(
            "Some error message raised in subclass"
        )
        self.assertEqual(0, MockJsonResponse.call_count)

    @patch("django_aws_api_gateway_websockets.views.HttpResponseBadRequest")
    @patch("django_aws_api_gateway_websockets.views.JsonResponse")
    def test_dispatch__with_missing_headers(
        self, MockJsonResponse, MockHttpResponseBadRequest
    ):
        """Should call the missing_headers method, returning it's result"""
        user = MagicMock(is_authenticated=True)
        request = self.factory.post(
            "?channel=my-channel",
            data=json.dumps({}),
            content_type="application/json",
            HTTP_Connectionid="1234",
            HTTP_Cookie="some-cookie",
            HTTP_Origin="example.com",
            HTTP_HOST="www.example.com",
            HTTP_ORIGIN="https://www.example.com",
        )
        request.user = user
        obj = views.WebSocketView()

        res = obj.dispatch(request)

        MockHttpResponseBadRequest.assert_called_with(
            "Some of the required headers are missing"
        )
        self.assertEqual(MockHttpResponseBadRequest.return_value, res)
        self.assertEqual(0, MockJsonResponse.call_count)

    @patch("django_aws_api_gateway_websockets.views.HttpResponseBadRequest")
    @patch("django_aws_api_gateway_websockets.views.JsonResponse")
    def test_dispatch__allowed_api_gateway_fails(
        self, MockJsonResponse, MockHttpResponseBadRequest
    ):
        """When _allowed_apigateway returns False the missing_headers method should be called"""
        user = MagicMock(is_authenticated=True)

        class SubClassedView(views.WebSocketView):
            def _allowed_apigateway(self, request, *args, **kwargs) -> bool:
                return False

            def missing_headers(self, request, *args, **kwargs):
                msg = "Got into missing headers"
                return self._return_bad_request(msg)

        request = self.factory.post(
            "?channel=my-channel",
            data=json.dumps({}),
            content_type="application/json",
            HTTP_Connectionid="1234",
            HTTP_Cookie="some-cookie",
            HTTP_Origin="example.com",
            HTTP_Sec_Websocket_Extensions="some-value",
            HTTP_Sec_Websocket_Key="some-key",
            HTTP_Sec_Websocket_Version="1.2.3",
            HTTP_HOST="www.example.com",
            HTTP_ORIGIN="https://www.example.com",
        )
        request.user = user
        obj = SubClassedView()

        res = obj.dispatch(request)

        MockHttpResponseBadRequest.assert_called_with("Got into missing headers")

        self.assertEqual(MockHttpResponseBadRequest.return_value, res)
        self.assertEqual(0, MockJsonResponse.call_count)

    @override_settings(ALLOWED_HOSTS=["www.example.com"])
    @patch("django_aws_api_gateway_websockets.views.WebSocketView._allowed_apigateway")
    @patch(
        "django_aws_api_gateway_websockets.views.WebSocketView.http_method_not_allowed"
    )
    @patch("django_aws_api_gateway_websockets.views.JsonResponse")
    def test_dispatch__when_http_method_is_not_allowed(
        self,
        MockJsonResponse,
        mocked_http_method_not_allowed,
        mocked__allowed_apigateway,
    ):
        """When a HTTP method is used that is not allowed, the http_method_not_allowed method should be invoked"""
        user = MagicMock(is_authenticated=True)
        mocked__allowed_apigateway.return_value = True

        class SubClassedView(views.WebSocketView):
            http_method_names = ["get"]

        request = self.factory.post(
            "?channel=my-channel",
            data=json.dumps({}),
            content_type="application/json",
            HTTP_HOST="www.example.com",
            HTTP_X_Forwarded_For="123.123.123.123",
            HTTP_X_Forwarded_Proto="https",
            HTTP_Content_Length=123,
            HTTP_Connectionid="1234",
            HTTP_User_Agent="Mozilla/5.0",
            HTTP_X_Amzn_Apigateway_Api_Id="APIGateway-1",
            HTTP_Connection="",
            HTTP_X_Amzn_Trace_Id="123456787654",
            HTTP_X_Forwarded_Port="443",
            HTTP_X_Real_Ip="172.0.0.1",
            HTTP_Cookie="some-cookie",
            HTTP_ORIGIN="https://www.example.com",
            HTTP_Sec_Websocket_Extensions="some-value",
            HTTP_Sec_Websocket_Key="some-key",
            HTTP_Sec_Websocket_Version="1.2.3",
        )
        request.user = user
        obj = SubClassedView()

        res = obj.dispatch(request)

        mocked_http_method_not_allowed.assert_called_with(request)
        self.assertEqual(0, MockJsonResponse.call_count)
        self.assertEqual(res, mocked_http_method_not_allowed.return_value, res.content)

    @override_settings(ALLOWED_HOSTS=["www.example.com"])
    @patch("django_aws_api_gateway_websockets.views.WebSocketView._allowed_apigateway")
    @patch("django_aws_api_gateway_websockets.views.WebSocketView.connect")
    @patch("django_aws_api_gateway_websockets.views.JsonResponse")
    def test_dispatch__route_is_connect(
        self, MockJsonResponse, mocked_connect, mocked__allowed_apigateway
    ):
        """When the route is connect and the headers, apigateway and method all pass checks the connect method should be invoked"""
        user = MagicMock(is_authenticated=True)
        mocked__allowed_apigateway.return_value = True

        request = self.factory.post(
            "",
            data=json.dumps({}),
            content_type="application/json",
            HTTP_HOST="www.example.com",
            HTTP_X_Forwarded_For="123.123.123.123",
            HTTP_X_Forwarded_Proto="https",
            HTTP_Content_Length=123,
            HTTP_Connectionid="1234",
            HTTP_User_Agent="Mozilla/5.0",
            HTTP_X_Amzn_Apigateway_Api_Id="APIGateway-1",
            HTTP_Connection="",
            HTTP_X_Amzn_Trace_Id="123456787654",
            HTTP_X_Forwarded_Port="443",
            HTTP_X_Real_Ip="172.0.0.1",
            HTTP_Cookie="some-cookie",
            HTTP_ORIGIN="https://www.example.com",
            HTTP_Sec_Websocket_Extensions="some-value",
            HTTP_Sec_Websocket_Key="some-key",
            HTTP_Sec_Websocket_Version="1.2.3",
        )
        request.user = user
        res = views.WebSocketView.as_view()(request, route="connect")

        mocked_connect.assert_called_with(request, route="connect")
        self.assertEqual(0, MockJsonResponse.call_count)
        self.assertEqual(res, mocked_connect.return_value, res)

    @override_settings(ALLOWED_HOSTS=["www.example.com"])
    @patch("django_aws_api_gateway_websockets.views.WebSocketView._allowed_apigateway")
    @patch("django_aws_api_gateway_websockets.views.WebSocketView._expected_useragent")
    @patch("django_aws_api_gateway_websockets.views.WebSocketView._add_user_to_request")
    @patch("django_aws_api_gateway_websockets.views.WebSocketView.disconnect")
    @patch("django_aws_api_gateway_websockets.views.JsonResponse")
    def test_dispatch__route_is_disconnect__when_useragent_is_expected(
        self,
        MockJsonResponse,
        mocked_disconnect,
        mocked__add_user_to_request,
        mocked__expected_useragent,
        mocked__allowed_apigateway,
    ):
        """When the route is connect and the headers, apigateway and method all pass checks the connect method should be invoked"""
        user = MagicMock(is_authenticated=True)
        mocked__allowed_apigateway.return_value = True
        mocked__expected_useragent.return_value = True
        mocked__add_user_to_request.return_value = True

        request = self.factory.post(
            "",
            data=json.dumps({}),
            content_type="application/json",
            HTTP_HOST="www.example.com",
            HTTP_X_Forwarded_For="123.123.123.123",
            HTTP_X_Forwarded_Proto="https",
            HTTP_Content_Length=123,
            HTTP_Connectionid="1234",
            HTTP_User_Agent="Mozilla/5.0",
            HTTP_X_Amzn_Apigateway_Api_Id="APIGateway-1",
            HTTP_Connection="",
            HTTP_X_Amzn_Trace_Id="123456787654",
            HTTP_X_Forwarded_Port="443",
            HTTP_X_Real_Ip="172.0.0.1",
            HTTP_Cookie="some-cookie",
            HTTP_ORIGIN="https://www.example.com",
            HTTP_Sec_Websocket_Extensions="some-value",
            HTTP_Sec_Websocket_Key="some-key",
            HTTP_Sec_Websocket_Version="1.2.3",
        )
        request.user = user
        res = views.WebSocketView.as_view()(request, route="disconnect")

        mocked_disconnect.assert_called_with(request, route="disconnect")
        mocked__expected_useragent.assert_called_with(request, route="disconnect")
        mocked__add_user_to_request.assert_called_with(request)
        self.assertEqual(0, MockJsonResponse.call_count)
        self.assertEqual(res, mocked_disconnect.return_value)

    @override_settings(ALLOWED_HOSTS=["www.example.com"])
    @patch("django_aws_api_gateway_websockets.views.WebSocketView._allowed_apigateway")
    @patch("django_aws_api_gateway_websockets.views.WebSocketView._expected_useragent")
    @patch("django_aws_api_gateway_websockets.views.WebSocketView._add_user_to_request")
    @patch("django_aws_api_gateway_websockets.views.WebSocketView.invalid_useragent")
    @patch("django_aws_api_gateway_websockets.views.WebSocketView.disconnect")
    @patch("django_aws_api_gateway_websockets.views.JsonResponse")
    def test_dispatch__route_is_disconnect__when_useragent_is_not_expected(
        self,
        MockJsonResponse,
        mocked_disconnect,
        mock_invalid_useragent,
        mocked__add_user_to_request,
        mocked__expected_useragent,
        mocked__allowed_apigateway,
    ):
        """When the route is connect and the headers, apigateway and method all pass checks the connect method should be invoked"""
        user = MagicMock(is_authenticated=True)
        mocked__allowed_apigateway.return_value = True
        mocked__expected_useragent.return_value = False

        request = self.factory.post(
            "",
            data=json.dumps({}),
            content_type="application/json",
            HTTP_HOST="www.example.com",
            HTTP_X_Forwarded_For="123.123.123.123",
            HTTP_X_Forwarded_Proto="https",
            HTTP_Content_Length=123,
            HTTP_Connectionid="1234",
            HTTP_User_Agent="Mozilla/5.0",
            HTTP_X_Amzn_Apigateway_Api_Id="APIGateway-1",
            HTTP_Connection="",
            HTTP_X_Amzn_Trace_Id="123456787654",
            HTTP_X_Forwarded_Port="443",
            HTTP_X_Real_Ip="172.0.0.1",
            HTTP_Cookie="some-cookie",
            HTTP_ORIGIN="https://www.example.com",
            HTTP_Sec_Websocket_Extensions="some-value",
            HTTP_Sec_Websocket_Key="some-key",
            HTTP_Sec_Websocket_Version="1.2.3",
        )
        request.user = user
        res = views.WebSocketView.as_view()(request, route="disconnect")

        mock_invalid_useragent.assert_called_with(request, route="disconnect")

        mocked__expected_useragent.assert_called_with(request, route="disconnect")
        self.assertEqual(0, mocked_disconnect.call_count)
        self.assertEqual(0, mocked__add_user_to_request.call_count)
        self.assertEqual(0, MockJsonResponse.call_count)
        self.assertEqual(res, mock_invalid_useragent.return_value)

    @override_settings(ALLOWED_HOSTS=["www.example.com"])
    @patch("django_aws_api_gateway_websockets.views.WebSocketView._allowed_apigateway")
    @patch("django_aws_api_gateway_websockets.views.WebSocketView._expected_useragent")
    @patch("django_aws_api_gateway_websockets.views.WebSocketView._add_user_to_request")
    @patch("django_aws_api_gateway_websockets.views.WebSocketView.invalid_useragent")
    @patch("django_aws_api_gateway_websockets.views.WebSocketView.disconnect")
    @patch("django_aws_api_gateway_websockets.views.WebSocketView._load_session")
    @patch("django_aws_api_gateway_websockets.views.JsonResponse")
    def test_dispatch__route_is_custom__useragent_is_expected(
        self,
        MockJsonResponse,
        mocked__load_session,
        mocked_disconnect,
        mock_invalid_useragent,
        mocked__add_user_to_request,
        mocked__expected_useragent,
        mocked__allowed_apigateway,
    ):
        """When the route is a custom one your method should be invoked"""
        user = MagicMock(is_authenticated=True)
        mocked__allowed_apigateway.return_value = True
        mocked__expected_useragent.return_value = True

        class SubClassedView(views.WebSocketView):
            def chat(self, request, *args, **kwargs):
                return "This is the chat method"

        request = self.factory.post(
            "",
            data=json.dumps({"action": "chat"}),
            content_type="application/json",
            HTTP_HOST="www.example.com",
            HTTP_X_Forwarded_For="123.123.123.123",
            HTTP_X_Forwarded_Proto="https",
            HTTP_Content_Length=123,
            HTTP_Connectionid="1234",
            HTTP_User_Agent="Mozilla/5.0",
            HTTP_X_Amzn_Apigateway_Api_Id="APIGateway-1",
            HTTP_Connection="",
            HTTP_X_Amzn_Trace_Id="123456787654",
            HTTP_X_Forwarded_Port="443",
            HTTP_X_Real_Ip="172.0.0.1",
            HTTP_Cookie="some-cookie",
            HTTP_ORIGIN="https://www.example.com",
            HTTP_Sec_Websocket_Extensions="some-value",
            HTTP_Sec_Websocket_Key="some-key",
            HTTP_Sec_Websocket_Version="1.2.3",
        )
        request.user = user
        res = SubClassedView.as_view()(request, route="chat")

        self.assertEqual(res, "This is the chat method")
        mocked__expected_useragent.assert_called_with(request, route="chat")
        mocked__add_user_to_request.assert_called_with(request)
        mocked__load_session.assert_called_with(request)
        self.assertEqual(0, mocked_disconnect.call_count)
        self.assertEqual(0, MockJsonResponse.call_count)
        self.assertEqual(0, mock_invalid_useragent.call_count)

    @override_settings(ALLOWED_HOSTS=["www.example.com"])
    @patch("django_aws_api_gateway_websockets.views.WebSocketView._allowed_apigateway")
    @patch("django_aws_api_gateway_websockets.views.WebSocketView._expected_useragent")
    @patch("django_aws_api_gateway_websockets.views.WebSocketView._add_user_to_request")
    @patch("django_aws_api_gateway_websockets.views.WebSocketView.invalid_useragent")
    @patch("django_aws_api_gateway_websockets.views.WebSocketView.disconnect")
    @patch("django_aws_api_gateway_websockets.views.WebSocketView._load_session")
    @patch("django_aws_api_gateway_websockets.views.JsonResponse")
    def test_dispatch__route_is_custom__but_useragent_is_unexpected(
        self,
        MockJsonResponse,
        mocked__load_session,
        mocked_disconnect,
        mock_invalid_useragent,
        mocked__add_user_to_request,
        mocked__expected_useragent,
        mocked__allowed_apigateway,
    ):
        """When the route is a custom one your method should be invoked"""
        user = MagicMock(is_authenticated=True)
        mocked__allowed_apigateway.return_value = True
        mocked__expected_useragent.return_value = False

        class SubClassedView(views.WebSocketView):
            def chat(self, request, *args, **kwargs):
                return "This is the chat method"

        request = self.factory.post(
            "",
            data=json.dumps({"action": "chat"}),
            content_type="application/json",
            HTTP_HOST="www.example.com",
            HTTP_X_Forwarded_For="123.123.123.123",
            HTTP_X_Forwarded_Proto="https",
            HTTP_Content_Length=123,
            HTTP_Connectionid="1234",
            HTTP_User_Agent="Mozilla/5.0",
            HTTP_X_Amzn_Apigateway_Api_Id="APIGateway-1",
            HTTP_Connection="",
            HTTP_X_Amzn_Trace_Id="123456787654",
            HTTP_X_Forwarded_Port="443",
            HTTP_X_Real_Ip="172.0.0.1",
            HTTP_Cookie="some-cookie",
            HTTP_ORIGIN="https://www.example.com",
            HTTP_Sec_Websocket_Extensions="some-value",
            HTTP_Sec_Websocket_Key="some-key",
            HTTP_Sec_Websocket_Version="1.2.3",
        )
        request.user = user
        res = SubClassedView.as_view()(request, route="chat")

        self.assertEqual(res, mock_invalid_useragent.return_value)

        mocked__load_session.assert_called_with(request)
        mocked__expected_useragent.assert_called_with(request, route="chat")
        mock_invalid_useragent.assert_called_with(request, route="chat")
        self.assertEqual(0, mocked_disconnect.call_count)
        self.assertEqual(0, MockJsonResponse.call_count)
        self.assertEqual(0, mocked__add_user_to_request.call_count)

    @override_settings(ALLOWED_HOSTS=["www.example.com"])
    @patch("django_aws_api_gateway_websockets.views.WebSocketView._allowed_apigateway")
    @patch("django_aws_api_gateway_websockets.views.WebSocketView._expected_useragent")
    @patch("django_aws_api_gateway_websockets.views.WebSocketView._add_user_to_request")
    @patch("django_aws_api_gateway_websockets.views.WebSocketView.invalid_useragent")
    @patch("django_aws_api_gateway_websockets.views.WebSocketView.disconnect")
    @patch("django_aws_api_gateway_websockets.views.WebSocketView._load_session")
    @patch(
        "django_aws_api_gateway_websockets.views.WebSocketView.route_selection_key_missing"
    )
    @patch("django_aws_api_gateway_websockets.views.JsonResponse")
    def test_dispatch__route_selection_key_missing(
        self,
        MockJsonResponse,
        mocked_route_selection_key_missing,
        mocked__load_session,
        mocked_disconnect,
        mock_invalid_useragent,
        mocked__add_user_to_request,
        mocked__expected_useragent,
        mocked__allowed_apigateway,
    ):
        """When the route is a custom one your method should be invoked"""
        user = MagicMock(is_authenticated=True)
        mocked__allowed_apigateway.return_value = True
        mocked__expected_useragent.return_value = False

        class SubClassedView(views.WebSocketView):
            def chat(self, request, *args, **kwargs):
                return "This is the chat method"

        request = self.factory.post(
            "",
            data=json.dumps({}),
            content_type="application/json",
            HTTP_HOST="www.example.com",
            HTTP_X_Forwarded_For="123.123.123.123",
            HTTP_X_Forwarded_Proto="https",
            HTTP_Content_Length=123,
            HTTP_Connectionid="1234",
            HTTP_User_Agent="Mozilla/5.0",
            HTTP_X_Amzn_Apigateway_Api_Id="APIGateway-1",
            HTTP_Connection="",
            HTTP_X_Amzn_Trace_Id="123456787654",
            HTTP_X_Forwarded_Port="443",
            HTTP_X_Real_Ip="172.0.0.1",
            HTTP_Cookie="some-cookie",
            HTTP_ORIGIN="https://www.example.com",
            HTTP_Sec_Websocket_Extensions="some-value",
            HTTP_Sec_Websocket_Key="some-key",
            HTTP_Sec_Websocket_Version="1.2.3",
        )
        request.user = user
        res = SubClassedView.as_view()(request, route="chat")

        self.assertEqual(res, mocked_route_selection_key_missing.return_value)

        mocked_route_selection_key_missing.assert_called_with(request, route="chat")

        self.assertEqual(0, mocked__load_session.call_count)
        self.assertEqual(0, mocked__expected_useragent.call_count)
        self.assertEqual(0, mock_invalid_useragent.call_count)
        self.assertEqual(0, mocked_disconnect.call_count)
        self.assertEqual(0, MockJsonResponse.call_count)
        self.assertEqual(0, mocked__add_user_to_request.call_count)



from django.core.management.base import BaseCommand

from django_aws_api_gateway_websockets.models import WebSocketSession


class Command(BaseCommand):
    """Removes the dead websocket connections from the table"""

    def handle(self, *args, **options):
        """Removed disconnected websockets"""
        WebSocketSession.objects.filter(connected=False).delete()



# Generated by Django 3.2.11 on 2022-12-08 12:27

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("django_aws_api_gateway_websockets", "0003_auto_20220706_1018"),
    ]

    operations = [
        migrations.AddField(
            model_name="apigateway",
            name="deployment_id",
            field=models.CharField(
                blank=True, default="", editable=False, max_length=32
            ),
        ),
        migrations.CreateModel(
            name="ApiGatewayAdditionalRoute",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "name",
                    models.CharField(
                        help_text="Descriptive name for the route", max_length=63
                    ),
                ),
                ("route_key", models.CharField(max_length=64, unique=True)),
                ("integration_url", models.URLField()),
                ("created_on", models.DateTimeField(auto_now_add=True)),
                ("updated_on", models.DateTimeField(auto_now=True)),
                (
                    "api_gateway",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="additional_routes",
                        to="django_aws_api_gateway_websockets.apigateway",
                    ),
                ),
            ],
        ),
    ]



import json

import boto3
from botocore.exceptions import ClientError
from django.conf import settings
from django.db import models


def get_boto3_client(service: str = "apigatewayv2", **kwargs):
    """Returns the boto3 client to use.

    :param str servivce: apigatewayv2 | apigatewaymanagementapi

    If you are using an IAM Role then you just need to set AWS_REGION_NAME within settings.py otherwise you need to
    set AWS_ACCESS_KEY_ID and AWS_SECRET_ACCESS_KEY as well with the correct values
    """
    if (
        hasattr(settings, "AWS_ACCESS_KEY_ID")
        and settings.AWS_ACCESS_KEY_ID
        and hasattr(settings, "AWS_SECRET_ACCESS_KEY")
        and settings.AWS_SECRET_ACCESS_KEY
    ):
        if not hasattr(settings, "AWS_REGION_NAME") or not settings.AWS_REGION_NAME:
            raise RuntimeError("AWS_REGION_NAME must be set within settings.py")

        client = boto3.client(
            service,
            aws_access_key_id=settings.AWS_ACCESS_KEY_ID,
            aws_secret_access_key=settings.AWS_SECRET_ACCESS_KEY,
            region_name=settings.AWS_REGION_NAME,
            **kwargs,
        )
    else:
        if not hasattr(settings, "AWS_REGION_NAME") or not settings.AWS_REGION_NAME:
            raise RuntimeError("AWS_REGION_NAME must be set within settings.py")
        client = boto3.client(service, region_name=settings.AWS_REGION_NAME, **kwargs)

    return client


class ApiGateway(models.Model):
    """Stored the API Gateway definitions"""

    class Meta:
        indexes = [
            models.Index(fields=["api_id"]),
            models.Index(fields=["domain_name"]),
            models.Index(fields=["custom_domain_created"]),
            models.Index(fields=["api_created"]),
            models.Index(fields=["created_on"]),
        ]

    def __str__(self) -> str:
        return self.api_name

    def save(self, **kwargs):
        """Ensure the trailing slash is saved to the target endpoint"""
        if not self.target_base_endpoint[-1:] == "/":
            self.target_base_endpoint = f"{self.target_base_endpoint}/"
        super().save(**kwargs)

    api_name = models.CharField(max_length=255, unique=True)
    api_description = models.CharField(max_length=255, blank=True, default="")
    default_channel_name = models.CharField(
        max_length=128,
        blank=True,
        default="",
        help_text="Automatically sets the 'channel' on the WebSocketSession record for the connection",
    )
    domain_name = models.CharField(
        max_length=255,
        blank=True,
        default="",
        help_text="The full domain you wish to use for the API endpoint. E.G ws.example.com",
    )
    target_base_endpoint = models.URLField(
        null=True,
        blank=True,
        default=None,
        help_text=(
            "The URL on your website where the API Gateway routes will point, including the trailing /, but excluding "
            "the final route/slug portion of the URL."
            " E.G. If your default route will point to https://www.example.com/ws/default then enter "
            "https://www.example.com/ws/"
        ),
    )
    certificate_arn = models.CharField(
        max_length=255,
        blank=True,
        default="",
        help_text="The ARN of the certificate to use from AWS Certificate Manager",
    )
    hosted_zone_id = models.CharField(
        max_length=32,
        blank=True,
        default="",
        help_text="The Hosted Zone ID from AWs Route 53 for the domain you wish to use",
    )
    api_key_selection_expression = models.CharField(
        max_length=255, default="$request.header.x-api-key"
    )
    route_selection_expression = models.CharField(
        max_length=255, default="$request.body.action"
    )
    route_key = models.CharField(max_length=255, default="$default")
    stage_name = models.CharField(max_length=63, default="production", blank=True)
    stage_description = models.CharField(max_length=63, default="", blank=True)
    deployment_id = models.CharField(
        max_length=32, default="", blank=True, editable=False
    )
    tags = models.JSONField(
        blank=True, default=dict, help_text='In format {"tag-name": "tag-value"}'
    )
    # Returned Values
    api_id = models.CharField(
        max_length=32,
        blank=True,
        default="",
        help_text="The ID of the Api Gateway returned by AWS",
    )
    api_endpoint = models.CharField(
        max_length=255, blank=True, default="", help_text="The Api Gateway endpoint"
    )
    api_gateway_domain_name = models.CharField(
        max_length=255,
        blank=True,
        default="",
        help_text="The value to point your CNAME record to",
    )
    api_mapping_id = models.CharField(
        max_length=128,
        blank=True,
        default="",
        help_text="The ApiMappingId to use with api_mapping calls",
    )
    api_created = models.BooleanField(default=False, editable=False)
    custom_domain_created = models.BooleanField(default=False, editable=False)
    created_on = models.DateTimeField(auto_now_add=True)
    updated_on = models.DateTimeField(auto_now=True)

    def create_gateway(self):
        """Creates the actual API gateway record"""
        if self.api_created:
            return

        client = get_boto3_client()
        self._create_api(client)
        try:
            self._create_routes(client)
            if self.pk:
                for additional_route in self.additional_routes.all():
                    if not additional_route.deployed:
                        additional_route.create_route(client, deploy=False)
            self._create_stage_and_deploy(client)
        except ClientError as ce:
            raise ce
        finally:
            self.api_created = True
            self.save()

    def create_custom_domain(self):
        """Uses boto3 to create the custom domain and associate it with the production stage of the loaded API

        Should be called after create_gateway() has been run
        :return:
        """
        if not self.api_created:
            raise ValueError("The API needs to be created before calling this method")
        if not self.certificate_arn:
            raise ValueError("A Certificate ARN is required")

        client = get_boto3_client()
        domain_res = self._create_domain_name(client)
        try:
            self.api_gateway_domain_name = domain_res["DomainNameConfigurations"][0][
                "ApiGatewayDomainName"
            ]
            self.api_mapping_id = self._create_api_mapping(client)
        except ClientError as ce:
            raise ce
        finally:
            self.custom_domain_created = True
            self.save()

    def _create_api(self, client):
        """Creates the base API Gateway endpoint"""
        res = client.create_api(
            ApiKeySelectionExpression=self.api_key_selection_expression,
            Description=self.api_description,
            DisableSchemaValidation=True,
            Name=self.api_name,
            ProtocolType="WEBSOCKET",
            RouteKey=self.route_key,
            RouteSelectionExpression=self.route_selection_expression,
            Target=f"{self.target_base_endpoint}/default",
        )
        self.api_id = res["ApiId"]
        self.api_endpoint = res["ApiEndpoint"]

    def _create_routes(self, client):
        """Creates the integrations and routes associating the route with integrations"""
        for route in ["$connect", "$disconnect", "$default"]:
            integration_res = client.create_integration(
                ApiId=self.api_id,
                ConnectionType="INTERNET",
                IntegrationMethod="POST",
                IntegrationType="HTTP_PROXY",
                IntegrationUri=f"{self.target_base_endpoint}{route.replace('$', '')}",
                PassthroughBehavior="WHEN_NO_MATCH",
                PayloadFormatVersion="1.0",
                RequestParameters={
                    "integration.request.header.connectionId": "context.connectionId"
                },
                TimeoutInMillis=29000,
            )
            extra_kwargs = {}
            if route == "$default":
                extra_kwargs["RouteResponseSelectionExpression"] = route
            client.create_route(
                ApiId=self.api_id,
                ApiKeyRequired=False,
                AuthorizationType="NONE",
                RouteKey=route,
                Target=f"integrations/{integration_res['IntegrationId']}",
                **extra_kwargs,
            )

    def _create_stage_and_deploy(self, client):
        """Create the stage and deployment"""
        client.create_stage(ApiId=self.api_id, StageName=self.stage_name)
        self.deploy_api(client)

    def deploy_api(self, client):
        res = client.create_deployment(
            ApiId=self.api_id,
            Description=self.stage_description,
            StageName=self.stage_name,
        )
        self.deployment_id = res["DeploymentId"]

    def _create_domain_name(self, client):
        """Creates the domain including the HostedZoneID if one is set"""
        conf = {
            "CertificateArn": self.certificate_arn,
            "DomainNameStatus": "AVAILABLE",
            "EndpointType": "REGIONAL",
            "SecurityPolicy": "TLS_1_2",
        }
        if self.hosted_zone_id:
            conf["HostedZoneId"] = self.hosted_zone_id

        return client.create_domain_name(
            DomainName=self.domain_name, DomainNameConfigurations=[conf]
        )

    def _create_api_mapping(self, client) -> str:
        mapping_res = client.create_api_mapping(
            ApiId=self.api_id, DomainName=self.domain_name, Stage=self.stage_name
        )
        return mapping_res["ApiMappingId"]


class ApiGatewayAdditionalRoute(models.Model):
    """Stores the additional route keys"""

    class Meta:
        # Once version 4 is the minimum supported vewrsion then swap to use this
        # constraints = [
        #     models.UniqueConstraint("api_gateway", 'name', name='unique_name_per_gateway')
        # ]
        unique_together = [["api_gateway", "route_key"]]

    def __str__(self) -> str:
        return self.name

    def save(self, *args, **kwargs):
        """Save the record then deploy the new route if the parent has already been deployed"""
        super().save(*args, **kwargs)
        if self.api_gateway.deployment_id and not self.deployed:
            client = get_boto3_client()
            self.create_route(client)

    api_gateway = models.ForeignKey(
        ApiGateway, on_delete=models.CASCADE, related_name="additional_routes"
    )
    name = models.CharField(max_length=63, help_text="Descriptive name for the route")
    route_key = models.CharField(max_length=64, db_index=True)
    integration_url = models.URLField()
    deployed = models.BooleanField(default=False, editable=False)
    created_on = models.DateTimeField(auto_now_add=True)
    updated_on = models.DateTimeField(auto_now=True)

    def create_route(self, client, deploy=True):
        """Create the Integration and then the route"""
        integration_res = client.create_integration(
            ApiId=self.api_gateway.api_id,
            ConnectionType="INTERNET",
            IntegrationMethod="POST",
            IntegrationType="HTTP_PROXY",
            IntegrationUri=self.integration_url,
            PassthroughBehavior="WHEN_NO_MATCH",
            PayloadFormatVersion="1.0",
            RequestParameters={
                "integration.request.header.connectionId": "context.connectionId"
            },
            TimeoutInMillis=29000,
        )
        client.create_route(
            ApiId=self.api_gateway.api_id,
            ApiKeyRequired=False,
            AuthorizationType="NONE",
            RouteKey=self.route_key,
            Target=f"integrations/{integration_res['IntegrationId']}",
            RouteResponseSelectionExpression="$default",
        )

        if deploy:
            self.api_gateway.deploy_api(client)
            self.deployed = True
            self.save()


class WebSocketSessionQuerySet(models.QuerySet):
    def send_message(self, data: dict):
        """Sends the same message to all WebSocketSessions included within the current filter

        Example use:

        Send a message to all active connections to the "Shared Channel Name" channel

        WebSocketSession.objects.filter(
            channel="Shared Channel Name"
        ).send_message(
            {
                "msg": "this is a server sent message"
            }
        )
        """
        client = None
        msg = json.dumps(data)
        res = []
        for obj in self.filter(connected=True):
            if not client:
                client = get_boto3_client(
                    "apigatewaymanagementapi",
                    endpoint_url=(
                        f"https://{obj.api_gateway.api_id}.execute-api."
                        f"{settings.AWS_REGION_NAME}.amazonaws.com/{obj.api_gateway.stage_name}"
                    ),
                )
            try:
                res.append(
                    client.post_to_connection(Data=msg, ConnectionId=obj.connection_id)
                )
            except ClientError as error:
                if error.response["Error"]["Code"] == "GoneException":
                    obj.connected = False
                    obj.save()
                else:
                    raise error

        return res


class WebSocketSession(models.Model):
    class Meta:
        indexes = [
            models.Index(fields=["connection_id", "connected"]),
            models.Index(fields=["channel_name", "connection_id"]),
            models.Index(fields=["channel_name", "connected"]),
        ]

    def __str__(self) -> str:
        return self.connection_id

    def send_message(self, data: dict):
        """Sends a message containing the given data to connection"""
        client = get_boto3_client(
            "apigatewaymanagementapi",
            endpoint_url=(
                f"https://{self.api_gateway.api_id}.execute-api."
                f"{settings.AWS_REGION_NAME}.amazonaws.com/{self.api_gateway.stage_name}"
            ),
        )
        try:
            return client.post_to_connection(
                Data=json.dumps(data), ConnectionId=self.connection_id
            )
        except ClientError as error:
            if error.response["Error"]["Code"] == "GoneException":
                self.connected = False
                self.save()
            else:
                raise error

    objects = WebSocketSessionQuerySet.as_manager()

    connection_id = models.CharField(max_length=255, unique=True)
    channel_name = models.CharField(
        max_length=128,
        blank=True,
        default="",
        help_text="Used to group connections together",
    )
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        blank=True,
        null=True,
        default=None,
        on_delete=models.CASCADE,
        related_name="websocket_sessions",
    )
    connected = models.BooleanField(
        default=True, help_text="Indicates is the connection is current or not"
    )
    api_gateway = models.ForeignKey(
        ApiGateway,
        null=True,
        blank=True,
        default=None,
        on_delete=models.SET_NULL,
        related_name="sessions",
    )
    request_count = models.PositiveBigIntegerField(default=1)
    created_on = models.DateTimeField(auto_now_add=True)
    updated_on = models.DateTimeField(auto_now=True)



from django.core.management.base import BaseCommand

from django_aws_api_gateway_websockets.models import ApiGateway


class Command(BaseCommand):
    def add_arguments(self, parser):
        """One required argument: pk which determine the ApiGateway object to run"""
        parser.add_argument(
            "--pk",
            required=True,
            type=int,
            help="The PK of the record you wish to run",
        )

    def handle(self, *args, **options):
        """Handles the request"""
        api = ApiGateway.objects.get(
            pk=options["pk"], api_created=True, custom_domain_created=False
        )
        api.create_custom_domain()



from unittest.mock import MagicMock, patch

from botocore import exceptions
from django.contrib.admin.sites import AdminSite
from django.contrib.auth.models import User
from django.test import RequestFactory, SimpleTestCase, TestCase

from django_aws_api_gateway_websockets import admin, models


class CreateApiGatewaySimpleTestCase(SimpleTestCase):
    @patch("django_aws_api_gateway_websockets.admin.messages")
    def test_empty_queryset(self, mocked_messages):
        """Empty queryset means nothing should run"""
        queryset = []

        admin.create_api_gateway(None, None, queryset)

        self.assertEqual(0, mocked_messages.success.call_count)
        self.assertEqual(0, mocked_messages.error.call_count)

    @patch("django_aws_api_gateway_websockets.admin.messages")
    def test_api_not_yet_created(self, mocked_messages):
        """When the record does not yet have an API that has been created, it should call create_gateway"""
        request = MagicMock()
        instance = MagicMock(
            api_created=False,
            custom_domain_created=False,
            api_name="Test API",
        )
        queryset = [instance]

        admin.create_api_gateway(None, request, queryset)

        instance.create_gateway.assert_called_with()
        mocked_messages.success.assert_called_with(
            request, f"API Gateway endpoint created for {instance.api_name}"
        )
        self.assertEqual(0, mocked_messages.error.call_count)

    @patch("django_aws_api_gateway_websockets.admin.messages")
    def test_api_already_created(self, mocked_messages):
        """When the api_created boolean is True the record should be skipped"""
        request = MagicMock()
        instance = MagicMock(
            api_created=True,
            custom_domain_created=False,
            api_name="Test API",
        )
        queryset = [instance]

        admin.create_api_gateway(None, request, queryset)

        self.assertEqual(0, instance.create_gateway.call_count)
        self.assertEqual(0, mocked_messages.success.call_count)
        self.assertEqual(0, mocked_messages.error.call_count)

    @patch("django_aws_api_gateway_websockets.admin.messages")
    def test_client_error_raised(self, mocked_messages):
        """When a ClientError occurs an error message should be added"""
        request = MagicMock()
        instance = MagicMock(
            api_created=False,
            custom_domain_created=False,
            api_name="Test API",
        )
        instance.create_gateway.side_effect = exceptions.ClientError(
            {"Error": {"Code": "A123", "Message": "Some error"}}, "apigateway"
        )
        queryset = [instance]

        admin.create_api_gateway(None, request, queryset)

        instance.create_gateway.assert_called_with()
        self.assertEqual(0, mocked_messages.success.call_count)
        mocked_messages.error.assert_called_with(
            request,
            (
                "An error occurred (A123) when calling the apigateway operation: Some error occurred when processing "
                f"{instance.api_name}"
            ),
        )


class CreatCustomDomainSimpleTestCase(SimpleTestCase):
    @patch("django_aws_api_gateway_websockets.admin.messages")
    def test_empty_queryset(self, mocked_messages):
        """Empty queryset means nothing should run"""
        queryset = []

        admin.create_custom_domain(None, None, queryset)

        self.assertEqual(0, mocked_messages.success.call_count)
        self.assertEqual(0, mocked_messages.error.call_count)

    @patch("django_aws_api_gateway_websockets.admin.messages")
    def test_api_created_but_custom_domain_not_created(self, mocked_messages):
        """When the record does not yet have an API that has been created, it should call create_gateway"""
        request = MagicMock()
        instance = MagicMock(
            api_created=True,
            custom_domain_created=False,
            api_name="Test API",
            api_gateway_domain_name="ws.example.com",
        )
        queryset = [instance]

        admin.create_custom_domain(None, request, queryset)

        instance.create_custom_domain.assert_called_with()
        mocked_messages.success.assert_called_with(
            request,
            f"Custom Domain setup for {instance.api_name}. Next update DNS to point to {instance.api_gateway_domain_name}",
        )
        self.assertEqual(0, mocked_messages.error.call_count)

    @patch("django_aws_api_gateway_websockets.admin.messages")
    def test_api_and_custom_domain_both_created(self, mocked_messages):
        """When the api_created boolean is True the record should be skipped"""
        request = MagicMock()
        instance = MagicMock(
            api_created=True,
            custom_domain_created=True,
            api_name="Test API",
            api_gateway_domain_name="ws.example.com",
        )
        queryset = [instance]

        admin.create_custom_domain(None, request, queryset)

        self.assertEqual(0, instance.create_custom_domain.call_count)
        self.assertEqual(0, mocked_messages.success.call_count)
        self.assertEqual(0, mocked_messages.error.call_count)

    @patch("django_aws_api_gateway_websockets.admin.messages")
    def test_client_error_raised(self, mocked_messages):
        """When a ClientError occurs an error message should be added"""
        request = MagicMock()
        instance = MagicMock(
            api_created=True,
            custom_domain_created=False,
            api_name="Test API",
            api_gateway_domain_name="ws.example.com",
        )
        instance.create_custom_domain.side_effect = exceptions.ClientError(
            {"Error": {"Code": "A123", "Message": "Some error"}}, "apigateway"
        )
        queryset = [instance]

        admin.create_custom_domain(None, request, queryset)

        instance.create_custom_domain.assert_called_with()
        self.assertEqual(0, mocked_messages.success.call_count)
        mocked_messages.error.assert_called_with(
            request,
            (
                "An error occurred (A123) when calling the apigateway operation: Some error occurred when processing "
                f"{instance.api_name}"
            ),
        )


class AdminTestCase(TestCase):
    """Tests that require the database"""

    def setUp(self) -> None:
        self.user, _ = User.objects.get_or_create(
            username="test", is_staff=True, is_superuser=True
        )

    def test_WebSocketSessionAdmin_queryset(self):
        """Ensure the overloaded queryset method is working"""
        site = admin.WebSocketSessionAdmin(
            model=models.WebSocketSession, admin_site=AdminSite()
        )
        request = RequestFactory().get("/view-path")
        site.get_queryset(request)

    def test_ApiGatewayAdmin_queryset(self):
        """Ensure the overloaded queryset method is working"""
        site = admin.ApiGatewayAdmin(model=models.ApiGateway, admin_site=AdminSite())
        request = RequestFactory().get("/view-path")
        site.get_queryset(request)

    def test_ApiGatewayAdditionalRouteAdmin_queryset(self):
        """Ensure the overloaded queryset method is working"""
        site = admin.ApiGatewayAdditionalRouteAdmin(
            model=models.ApiGatewayAdditionalRoute, admin_site=AdminSite()
        )
        request = RequestFactory().get("/view-path")
        site.get_queryset(request)

    def test_ApiGatewayAdditionalRouteInline_queryset(self):
        """Ensure the overloaded queryset method is working"""
        site = admin.ApiGatewayAdditionalRouteInline(
            parent_model=models.ApiGatewayAdditionalRoute, admin_site=AdminSite()
        )
        request = RequestFactory().get("/view-path")
        request.user = self.user
        request.session = {}
        site.get_queryset(request)



from django.apps import AppConfig
from django.utils.translation import gettext_lazy as _


class DjangoAwsApiGatewayWebsocketsConfig(AppConfig):
    name = "django_aws_api_gateway_websockets"
    default_auto_field = "django.db.models.BigAutoField"
    verbose_name = _("Django AWS ApiGateway Websockets")



# Generated by Django 3.2.11 on 2022-12-15 08:47

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        (
            "django_aws_api_gateway_websockets",
            "0005_apigatewayadditionalroute_deployed",
        ),
    ]

    operations = [
        migrations.AlterField(
            model_name="apigatewayadditionalroute",
            name="route_key",
            field=models.CharField(db_index=True, max_length=64),
        ),
        migrations.AlterUniqueTogether(
            name="apigatewayadditionalroute",
            unique_together={("api_gateway", "route_key")},
        ),
    ]



# Generated by Django 3.2.11 on 2022-01-17 15:23

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="WebSocketSession",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("connection_id", models.CharField(max_length=255, unique=True)),
                (
                    "channel_name",
                    models.CharField(
                        blank=True,
                        default="",
                        help_text="Used to group connections together",
                        max_length=128,
                    ),
                ),
                (
                    "connected",
                    models.BooleanField(
                        default=True,
                        help_text="Indicates is the connection is current or not",
                    ),
                ),
                ("request_count", models.PositiveBigIntegerField(default=1)),
                ("created_on", models.DateTimeField(auto_now_add=True)),
                ("updated_on", models.DateTimeField(auto_now=True)),
                (
                    "author",
                    models.ForeignKey(
                        blank=True,
                        default=None,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="websocket_sessions",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
        ),
        migrations.AddIndex(
            model_name="websocketsession",
            index=models.Index(
                fields=["connection_id", "connected"],
                name="django_aws__connect_fee1eb_idx",
            ),
        ),
        migrations.AddIndex(
            model_name="websocketsession",
            index=models.Index(
                fields=["channel_name", "connection_id"],
                name="django_aws__channel_dd10fa_idx",
            ),
        ),
        migrations.AddIndex(
            model_name="websocketsession",
            index=models.Index(
                fields=["channel_name", "connected"],
                name="django_aws__channel_1de026_idx",
            ),
        ),
    ]



# Generated by Django 3.2.11 on 2022-12-08 13:37

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("django_aws_api_gateway_websockets", "0004_auto_20221208_1227"),
    ]

    operations = [
        migrations.AddField(
            model_name="apigatewayadditionalroute",
            name="deployed",
            field=models.BooleanField(default=False, editable=False),
        ),
    ]



import json
from unittest.mock import MagicMock, call, patch

from botocore.exceptions import ClientError
from django.conf import settings
from django.test import SimpleTestCase, TestCase, override_settings

from django_aws_api_gateway_websockets.models import (
    ApiGateway,
    ApiGatewayAdditionalRoute,
    WebSocketSession,
    get_boto3_client,
)


class GetBoto3ClientTestCase(SimpleTestCase):
    @override_settings(
        AWS_IAM_PROFILE="FakeIAMProfile",
        AWS_ACCESS_KEY_ID=None,
        AWS_SECRET_ACCESS_KEY=None,
        AWS_REGION_NAME=None,
    )
    @patch("django_aws_api_gateway_websockets.models.boto3")
    def test_exception_raised_if_no_credentials_other_than_aws_iam_profile(
        self, mock_boto3
    ):
        """Named profiles are not supported yet by this package. In theory they may be the preferred method for
         local and non-AWS hosted deployments.

         todo - I will add support in later

        :param MagicMock mock_boto3:
        :return:
        """
        with self.assertRaises(RuntimeError) as e:
            get_boto3_client("s3")

        self.assertEqual(
            str(e.exception), "AWS_REGION_NAME must be set within settings.py"
        )

    @override_settings(
        AWS_ACCESS_KEY_ID=None,
        AWS_SECRET_ACCESS_KEY=None,
        AWS_REGION_NAME="eu-west-1",
    )
    @patch("django_aws_api_gateway_websockets.models.boto3")
    def test_connection_made_if_aws_region_name_given_but_no_other_credentials(
        self, mock_boto3
    ):
        """This would be an example of the project running on an EC2 instance with a IAM profile assigned to it

        :param MagicMock mock_boto3:
        :return:
        """
        res = get_boto3_client("s3")
        mock_boto3.client.assert_called_with("s3", region_name="eu-west-1")

        self.assertEqual(res, mock_boto3.client.return_value)

    @override_settings(
        AWS_ACCESS_KEY_ID=None,
        AWS_SECRET_ACCESS_KEY=None,
        AWS_REGION_NAME="eu-west-1",
    )
    @patch("django_aws_api_gateway_websockets.models.boto3")
    def test_additional_kwargs_are_passed_into_the_connection(self, mock_boto3):
        """Any additional kwargs passed into the function should be passed into the boto.client invocation

        :param MagicMock mock_boto3:
        :return:
        """
        res = get_boto3_client("s3", another="param")
        mock_boto3.client.assert_called_with(
            "s3", region_name="eu-west-1", another="param"
        )

        self.assertEqual(res, mock_boto3.client.return_value)

    @override_settings(
        AWS_ACCESS_KEY_ID="my-access-key",
        AWS_SECRET_ACCESS_KEY="my-secret-key",
        AWS_REGION_NAME="eu-west-1",
    )
    @patch("django_aws_api_gateway_websockets.models.boto3")
    def test_access_key_and_secret_key_used(self, mock_boto3):
        """If the Access Key and Secret Key are defined within settings.py then they will be used. The region name
        must be also set

        :param MagicMock mock_boto3:
        :return:
        """
        res = get_boto3_client("s3")
        mock_boto3.client.assert_called_with(
            "s3",
            aws_access_key_id=settings.AWS_ACCESS_KEY_ID,
            aws_secret_access_key=settings.AWS_SECRET_ACCESS_KEY,
            region_name=settings.AWS_REGION_NAME,
        )

        self.assertEqual(res, mock_boto3.client.return_value)

    @override_settings(
        AWS_ACCESS_KEY_ID="my-access-key",
        AWS_SECRET_ACCESS_KEY="my-secret-key",
        AWS_REGION_NAME="eu-west-1",
    )
    @patch("django_aws_api_gateway_websockets.models.boto3")
    def test_service_defaults_to_apigatewayv2(self, mock_boto3):
        """If the service is not passed into the get_boto_client function then it should default to apigatewayv2

        :param MagicMock mock_boto3:
        :return:
        """
        res = get_boto3_client()
        mock_boto3.client.assert_called_with(
            "apigatewayv2",
            aws_access_key_id=settings.AWS_ACCESS_KEY_ID,
            aws_secret_access_key=settings.AWS_SECRET_ACCESS_KEY,
            region_name=settings.AWS_REGION_NAME,
        )

        self.assertEqual(res, mock_boto3.client.return_value)

    @override_settings(
        AWS_ACCESS_KEY_ID="my-access-key",
        AWS_SECRET_ACCESS_KEY="my-secret-key",
        AWS_REGION_NAME="",
    )
    @patch("django_aws_api_gateway_websockets.models.boto3")
    def test_region_name_is_required(self, mock_boto3):
        """If the Access Key and Secret Key are defined within settings.py  but the region is not set then an exception
        should be raised

        :param MagicMock mock_boto3:
        :return:
        """
        with self.assertRaises(RuntimeError) as e:
            get_boto3_client("s3")

        self.assertEqual(
            str(e.exception), "AWS_REGION_NAME must be set within settings.py"
        )


class ApiGatewaySimpleTestCase(SimpleTestCase):
    def test__str___returns_api_name(self):
        """The __str__ method should return the value of api_name"""
        self.assertEqual("My Api", str(ApiGateway(api_name="My Api")))

    @patch("django_aws_api_gateway_websockets.models.get_boto3_client")
    def test_create_gateway__when_api_already_created(self, mocked_get_boto3_client):
        """Should return None without trying to create the API"""
        obj = ApiGateway(api_name="My Api", api_created=True)

        self.assertIsNone(obj.create_gateway())
        self.assertEqual(0, mocked_get_boto3_client.call_count)

    @patch("django_aws_api_gateway_websockets.models.get_boto3_client")
    @patch.object(ApiGateway, "_create_api")
    @patch.object(ApiGateway, "_create_stage_and_deploy")
    @patch.object(ApiGateway, "_create_routes")
    @patch.object(ApiGateway, "additional_routes")
    @patch.object(ApiGateway, "save")
    def test_create_gateway__no_errors(
        self,
        mocked_save,
        mocked_additional_routes,
        mocked__create_routes,
        mocked__create_stage_and_deploy,
        mocked___create_api,
        mocked_get_boto3_client,
    ):
        """When there are no errors the _create_api, _create_routes, _create_stage_and_deploy methods should be called"""
        obj = ApiGateway(pk=12, api_name="My Api", api_created=False)
        route_1 = ApiGatewayAdditionalRoute(deployed=True)
        route_2 = ApiGatewayAdditionalRoute(
            deployed=False, api_gateway=obj, integration_url="https://example.com/foo"
        )

        mocked_additional_routes.all.return_value = [route_1, route_2]
        obj = ApiGateway(pk=12, api_name="My Api", api_created=False)

        self.assertFalse(obj.api_created)
        obj.create_gateway()

        self.assertTrue(obj.api_created)
        mocked_get_boto3_client.assert_called_with()
        mocked___create_api.assert_called_with(mocked_get_boto3_client.return_value)
        mocked__create_routes.assert_called_with(mocked_get_boto3_client.return_value)
        mocked_additional_routes.all.assert_called_with()
        mocked_get_boto3_client.return_value.create_integration.assert_called_with(
            ApiId=route_2.api_gateway.api_id,
            ConnectionType="INTERNET",
            IntegrationMethod="POST",
            IntegrationType="HTTP_PROXY",
            IntegrationUri=route_2.integration_url,
            PassthroughBehavior="WHEN_NO_MATCH",
            PayloadFormatVersion="1.0",
            RequestParameters={
                "integration.request.header.connectionId": "context.connectionId"
            },
            TimeoutInMillis=29000,
        )
        mocked__create_stage_and_deploy.assert_called_with(
            mocked_get_boto3_client.return_value
        )
        mocked_save.assert_called_with()

    @patch("django_aws_api_gateway_websockets.models.get_boto3_client")
    @patch.object(ApiGateway, "_create_api")
    @patch.object(ApiGateway, "_create_routes")
    @patch.object(ApiGateway, "_create_stage_and_deploy")
    @patch.object(ApiGateway, "save")
    def test_create_gateway__when_create_api_raises_error(
        self,
        mocked_save,
        mocked__create_stage_and_deploy,
        mocked__create_routes,
        mocked___create_api,
        mocked_get_boto3_client,
    ):
        """When an exception is called during create_api the other methods should not be called and the api_created
        property should be left as False
        """
        obj = ApiGateway(api_name="My Api", api_created=False)

        mocked___create_api.side_effect = Exception("Unexpected Error")

        with self.assertRaises(Exception) as e:
            self.assertFalse(obj.api_created)
            obj.create_gateway()

        self.assertEqual("Unexpected Error", str(e.exception))

        mocked_get_boto3_client.assert_called_with()
        mocked___create_api.assert_called_with(mocked_get_boto3_client.return_value)
        self.assertFalse(obj.api_created)
        self.assertEqual(0, mocked__create_routes.call_count)
        self.assertEqual(0, mocked__create_stage_and_deploy.call_count)
        self.assertEqual(0, mocked_save.call_count)

    @patch("django_aws_api_gateway_websockets.models.get_boto3_client")
    @patch.object(ApiGateway, "_create_api")
    @patch.object(ApiGateway, "_create_routes")
    @patch.object(ApiGateway, "_create_stage_and_deploy")
    @patch.object(ApiGateway, "save")
    def test_create_gateway__when__create_routes_raises_clienterror(
        self,
        mocked_save,
        mocked__create_stage_and_deploy,
        mocked__create_routes,
        mocked___create_api,
        mocked_get_boto3_client,
    ):
        """When an exception is called during _create_routes the other methods should not be called BUT the api_created
        property should be sert to as True
        """
        obj = ApiGateway(api_name="My Api", api_created=False)

        mocked__create_routes.side_effect = ClientError(
            {"Error": {"Code": "ABC123", "Message": "Actual Error Here"}}, "apigateway"
        )
        with self.assertRaises(Exception) as e:
            self.assertFalse(obj.api_created)
            obj.create_gateway()

        self.assertTrue(obj.api_created)
        mocked_save.assert_called_with()
        self.assertEqual(
            "An error occurred (ABC123) when calling the apigateway operation: Actual Error Here",
            str(e.exception),
        )

        mocked_get_boto3_client.assert_called_with()
        mocked___create_api.assert_called_with(mocked_get_boto3_client.return_value)
        mocked__create_routes.assert_called_with(mocked_get_boto3_client.return_value)
        self.assertEqual(0, mocked__create_stage_and_deploy.call_count)

    @patch("django_aws_api_gateway_websockets.models.get_boto3_client")
    @patch.object(ApiGateway, "_create_api")
    @patch.object(ApiGateway, "_create_routes")
    @patch.object(ApiGateway, "_create_stage_and_deploy")
    @patch.object(ApiGateway, "save")
    def test_create_gateway__when__create_stage_and_deploy_clienterror(
        self,
        mocked_save,
        mocked__create_stage_and_deploy,
        mocked__create_routes,
        mocked___create_api,
        mocked_get_boto3_client,
    ):
        """When an exception is called during _create_stage_and_deploy the api_created property should be set to as
        True
        """
        obj = ApiGateway(api_name="My Api", api_created=False)

        mocked__create_stage_and_deploy.side_effect = ClientError(
            {"Error": {"Code": "ABC123", "Message": "Actual Error Here"}}, "apigateway"
        )
        with self.assertRaises(Exception) as e:
            self.assertFalse(obj.api_created)
            obj.create_gateway()

        self.assertTrue(obj.api_created)
        mocked_save.assert_called_with()
        self.assertEqual(
            "An error occurred (ABC123) when calling the apigateway operation: Actual Error Here",
            str(e.exception),
        )

        mocked_get_boto3_client.assert_called_with()
        mocked___create_api.assert_called_with(mocked_get_boto3_client.return_value)
        mocked__create_routes.assert_called_with(mocked_get_boto3_client.return_value)
        mocked__create_stage_and_deploy.assert_called_with(
            mocked_get_boto3_client.return_value
        )

    @patch("django_aws_api_gateway_websockets.models.get_boto3_client")
    @patch.object(ApiGateway, "_create_domain_name")
    @patch.object(ApiGateway, "_create_api_mapping")
    @patch.object(ApiGateway, "save")
    def test_create_custom_domain__raises_valueerror_when_api_created_is_false(
        self,
        mocked_save,
        mocked__create_api_mapping,
        mocked__create_domain_name,
        mocked_get_boto3_client,
    ):
        """When api_created is False then a ValueError should be raised"""
        obj = ApiGateway(api_name="My Api", api_created=False)

        with self.assertRaises(Exception) as e:
            obj.create_custom_domain()

        self.assertEqual(
            str(e.exception), "The API needs to be created before calling this method"
        )

        self.assertEqual(0, mocked_get_boto3_client.call_count)
        self.assertEqual(0, mocked__create_domain_name.call_count)
        self.assertEqual(0, mocked__create_api_mapping.call_count)
        self.assertEqual(0, mocked_save.call_count)

    @patch("django_aws_api_gateway_websockets.models.get_boto3_client")
    @patch.object(ApiGateway, "_create_domain_name")
    @patch.object(ApiGateway, "_create_api_mapping")
    @patch.object(ApiGateway, "save")
    def test_create_custom_domain__raises_valueerror_when_certificate_arn_is_not_set(
        self,
        mocked_save,
        mocked__create_api_mapping,
        mocked__create_domain_name,
        mocked_get_boto3_client,
    ):
        """When api_created is True but the certificate_arn is not set then a ValueError should be raised"""
        obj = ApiGateway(api_name="My Api", api_created=True, certificate_arn="")

        with self.assertRaises(Exception) as e:
            obj.create_custom_domain()

        self.assertEqual(str(e.exception), "A Certificate ARN is required")

        self.assertEqual(0, mocked_get_boto3_client.call_count)
        self.assertEqual(0, mocked__create_domain_name.call_count)
        self.assertEqual(0, mocked__create_api_mapping.call_count)
        self.assertEqual(0, mocked_save.call_count)

    @patch("django_aws_api_gateway_websockets.models.get_boto3_client")
    @patch.object(ApiGateway, "_create_domain_name")
    @patch.object(ApiGateway, "_create_api_mapping")
    @patch.object(ApiGateway, "save")
    def test_create_custom_domain__as_expected(
        self,
        mocked_save,
        mocked__create_api_mapping,
        mocked__create_domain_name,
        mocked_get_boto3_client,
    ):
        """When api_created is False, the certificate arn is set and no errors are raised"""
        domain_result = {
            "DomainNameConfigurations": [
                {"ApiGatewayDomainName": "api-gateway.websocket.example.com"}
            ]
        }
        mocked__create_domain_name.return_value = domain_result

        obj = ApiGateway(
            api_name="My Api",
            api_created=True,
            certificate_arn="aws123:zxcv:3455",
            custom_domain_created=False,
            api_gateway_domain_name="",
            api_mapping_id="",
        )

        self.assertFalse(obj.custom_domain_created)

        obj.create_custom_domain()

        mocked_get_boto3_client.assert_called_with()
        mocked__create_domain_name.assert_called_with(
            mocked_get_boto3_client.return_value
        )
        mocked__create_api_mapping.asset_called_with(
            mocked_get_boto3_client.return_value
        )
        mocked_save.assert_called_with()

        self.assertTrue(obj.custom_domain_created)
        self.assertEqual(
            "api-gateway.websocket.example.com", obj.api_gateway_domain_name
        )
        self.assertEqual(obj.api_mapping_id, mocked__create_api_mapping.return_value)

    @patch("django_aws_api_gateway_websockets.models.get_boto3_client")
    @patch.object(ApiGateway, "_create_domain_name")
    @patch.object(ApiGateway, "_create_api_mapping")
    @patch.object(ApiGateway, "save")
    def test_create_custom_domain__clienterror_raised(
        self,
        mocked_save,
        mocked__create_api_mapping,
        mocked__create_domain_name,
        mocked_get_boto3_client,
    ):
        """When api_created is False, the certificate arn is set and no errors are raised"""
        mocked__create_api_mapping.side_effect = ClientError(
            {"Error": {"Code": "ABC123", "Message": "Actual Error Here"}}, "apigateway"
        )

        domain_result = {
            "DomainNameConfigurations": [
                {"ApiGatewayDomainName": "api-gateway.websocket.example.com"}
            ]
        }
        mocked__create_domain_name.return_value = domain_result

        obj = ApiGateway(
            api_name="My Api",
            api_created=True,
            certificate_arn="aws123:zxcv:3455",
            custom_domain_created=False,
            api_gateway_domain_name="",
            api_mapping_id="",
        )

        self.assertFalse(obj.custom_domain_created)

        with self.assertRaises(ClientError) as ce:
            obj.create_custom_domain()

        mocked_get_boto3_client.assert_called_with()
        mocked__create_domain_name.assert_called_with(
            mocked_get_boto3_client.return_value
        )
        mocked__create_api_mapping.asset_called_with(
            mocked_get_boto3_client.return_value
        )
        mocked_save.assert_called_with()

        self.assertTrue(obj.custom_domain_created)
        self.assertEqual(
            "api-gateway.websocket.example.com", obj.api_gateway_domain_name
        )
        self.assertEqual(
            str(ce.exception),
            "An error occurred (ABC123) when calling the apigateway operation: Actual Error Here",
        )

    def test__create_api(self):
        """When api_created is False, the certificate arn is set and no errors are raised"""
        client = MagicMock()
        client.create_api.return_value = {
            "ApiId": "1234ABCD",
            "ApiEndpoint": "https://api.amazon.com/this/url",
        }

        obj = ApiGateway(
            api_id="",
            api_endpoint="",
            api_key_selection_expression="$request.header.x-api-key",
            api_description="Some of description",
            api_name="",
            route_key="$default",
            route_selection_expression="$request.body.action",
            target_base_endpoint="https://example.com",
        )
        obj._create_api(client)

        client.create_api.assert_called_with(
            ApiKeySelectionExpression=obj.api_key_selection_expression,
            Description=obj.api_description,
            DisableSchemaValidation=True,
            Name=obj.api_name,
            ProtocolType="WEBSOCKET",
            RouteKey=obj.route_key,
            RouteSelectionExpression=obj.route_selection_expression,
            Target=f"{obj.target_base_endpoint}/default",
        )

        self.assertEqual("1234ABCD", obj.api_id)
        self.assertEqual("https://api.amazon.com/this/url", obj.api_endpoint)

    def test__create_routes(self):
        """create_integration and create_stage are should be called three times"""
        client = MagicMock()

        client.create_integration.side_effect = [
            {"IntegrationId": "i123-d345"},
            {"IntegrationId": "test-1234"},
            {"IntegrationId": "0997-tyty"},
        ]

        obj = ApiGateway(
            api_id="123456",
            api_endpoint="",
            api_key_selection_expression="$request.header.x-api-key",
            api_description="Some of description",
            api_name="",
            route_key="$default",
            route_selection_expression="$request.body.action",
            target_base_endpoint="https://example.com",
        )

        obj._create_routes(client)

        client.create_integration.assert_has_calls(
            [
                call(
                    ApiId=obj.api_id,
                    ConnectionType="INTERNET",
                    IntegrationMethod="POST",
                    IntegrationType="HTTP_PROXY",
                    IntegrationUri=f"{obj.target_base_endpoint}connect",
                    PassthroughBehavior="WHEN_NO_MATCH",
                    PayloadFormatVersion="1.0",
                    RequestParameters={
                        "integration.request.header.connectionId": "context.connectionId"
                    },
                    TimeoutInMillis=29000,
                ),
                call(
                    ApiId=obj.api_id,
                    ConnectionType="INTERNET",
                    IntegrationMethod="POST",
                    IntegrationType="HTTP_PROXY",
                    IntegrationUri=f"{obj.target_base_endpoint}disconnect",
                    PassthroughBehavior="WHEN_NO_MATCH",
                    PayloadFormatVersion="1.0",
                    RequestParameters={
                        "integration.request.header.connectionId": "context.connectionId"
                    },
                    TimeoutInMillis=29000,
                ),
                call(
                    ApiId=obj.api_id,
                    ConnectionType="INTERNET",
                    IntegrationMethod="POST",
                    IntegrationType="HTTP_PROXY",
                    IntegrationUri=f"{obj.target_base_endpoint}default",
                    PassthroughBehavior="WHEN_NO_MATCH",
                    PayloadFormatVersion="1.0",
                    RequestParameters={
                        "integration.request.header.connectionId": "context.connectionId"
                    },
                    TimeoutInMillis=29000,
                ),
            ]
        )

        client.create_route.assert_has_calls(
            [
                call(
                    ApiId=obj.api_id,
                    ApiKeyRequired=False,
                    AuthorizationType="NONE",
                    RouteKey="$connect",
                    Target=f"integrations/i123-d345",
                ),
                call(
                    ApiId=obj.api_id,
                    ApiKeyRequired=False,
                    AuthorizationType="NONE",
                    RouteKey="$disconnect",
                    Target=f"integrations/test-1234",
                ),
                call(
                    ApiId=obj.api_id,
                    ApiKeyRequired=False,
                    AuthorizationType="NONE",
                    RouteKey="$default",
                    Target=f"integrations/0997-tyty",
                    RouteResponseSelectionExpression="$default",
                ),
            ]
        )

    @patch("django_aws_api_gateway_websockets.models.get_boto3_client")
    def test__create_stage_and_deploy(self, mocked_get_boto3_client):
        """Should return None without trying to create the API"""
        client = MagicMock()
        obj = ApiGateway(
            api_id="1234",
            api_name="My Api",
            stage_name="My Stage Name",
            stage_description="Some description",
        )
        obj._create_stage_and_deploy(client)

        client.create_stage.assert_called_with(
            ApiId=obj.api_id, StageName=obj.stage_name
        )
        client.create_deployment.assert_called_with(
            ApiId=obj.api_id,
            Description=obj.stage_description,
            StageName=obj.stage_name,
        )

    def test__create_domain_name__no_hosted_zone_id(self):
        """Test the parameters passed into the Boto3 client when the HostedZoneID is not set"""
        client = MagicMock()
        obj = ApiGateway(
            domain_name="example.com",
            certificate_arn="arn:aws:acm:region:account:certificate/certificate_ID",
            hosted_zone_id="",
        )
        obj._create_domain_name(client)

        client.create_domain_name.assert_called_with(
            DomainName=obj.domain_name,
            DomainNameConfigurations=[
                {
                    "CertificateArn": obj.certificate_arn,
                    "DomainNameStatus": "AVAILABLE",
                    "EndpointType": "REGIONAL",
                    "SecurityPolicy": "TLS_1_2",
                }
            ],
        )

    def test__create_domain_name__with_hosted_zone_id(self):
        """Test the parameters passed into the Boto3 client when the HostedZoneID is set"""
        client = MagicMock()
        obj = ApiGateway(
            domain_name="example.com",
            certificate_arn="arn:aws:acm:region:account:certificate/certificate_ID",
            hosted_zone_id="ABCD-123",
        )
        obj._create_domain_name(client)

        client.create_domain_name.assert_called_with(
            DomainName=obj.domain_name,
            DomainNameConfigurations=[
                {
                    "CertificateArn": obj.certificate_arn,
                    "DomainNameStatus": "AVAILABLE",
                    "EndpointType": "REGIONAL",
                    "SecurityPolicy": "TLS_1_2",
                    "HostedZoneId": obj.hosted_zone_id,
                }
            ],
        )

    def test__create_api_mapping(self):
        """Test the parameters passed into the Boto3 client when the HostedZoneID is set"""
        client = MagicMock()
        client.create_api_mapping.return_value = {"ApiMappingId": "MyID"}

        obj = ApiGateway(
            api_id=1234, domain_name="example.com", stage_name="development"
        )

        res = obj._create_api_mapping(client)

        client.create_api_mapping.assert_called_with(
            ApiId=obj.api_id, DomainName=obj.domain_name, Stage=obj.stage_name
        )
        self.assertEqual("MyID", res)


class ApiGatewayIntegrationTest(TestCase):
    def test_save_ensures_endpoint_ends_with_forward_slash(self):
        """The target_base_endpoint should also end with a trailing slash. If it is not set then it should be set by
        the save method
        """
        endpoints = [
            {
                "name": "test 1",
                "use": "http://www.example1.com/something",
                "result": "http://www.example1.com/something/",
            },
            {
                "name": "test 2",
                "use": "http://www.example1.com/something/",
                "result": "http://www.example1.com/something/",
            },
            {
                "name": "test 3",
                "use": "http://www.example1.com",
                "result": "http://www.example1.com/",
            },
            {
                "name": "test 4",
                "use": "http://www.example1.com/",
                "result": "http://www.example1.com/",
            },
        ]

        for config in endpoints:
            with self.subTest(config=config):
                api = ApiGateway(
                    api_name=config["name"],
                    api_description="A test api",
                    target_base_endpoint=config["use"],
                )
                api.save()
                self.assertEqual(
                    api.target_base_endpoint, config["result"], config["name"]
                )


class ApiGatewayAdditionalRouteSimpleTestCase(SimpleTestCase):
    def test__str__returns_name(self):
        """The __str__ of ApiGatewayAdditionalRoute should return the name attribute"""
        obj = ApiGatewayAdditionalRoute(name="My Test")
        self.assertEqual("My Test", str(obj))


class ApiGatewayAdditionalRouteIntegrationTestCase(TestCase):
    def setUp(self) -> None:
        self.non_deployed_api_gateway = ApiGateway(
            api_name="First Gateway",
            api_description="A test api gateway",
            target_base_endpoint="http://www.example1.com/",
            deployment_id="",
        )
        self.non_deployed_api_gateway.save()

        self.deployed_api_gateway = ApiGateway(
            api_name="Second Gateway",
            api_description="A test api gateway",
            target_base_endpoint="http://www.example.com/ws/",
            deployment_id="ABC123",
        )
        self.deployed_api_gateway.save()

    @patch("django_aws_api_gateway_websockets.models.get_boto3_client")
    def test_adding_route_to_gateway_with_no_deployment_id(
        self, mocked_get_boto3_client
    ):
        """When adding a route to a non-deployed gateway no post-save logic should run"""
        ApiGatewayAdditionalRoute.objects.create(
            api_gateway=self.non_deployed_api_gateway,
            name="Alternative Route",
            route_key="another-route",
            integration_url="https://www.example.com/ws/other",
            deployed=False,
        )
        self.assertEqual(0, mocked_get_boto3_client.call_count)

    @patch("django_aws_api_gateway_websockets.models.get_boto3_client")
    def test_adding_route_to_gateway_with_no_deployment_id_but_deployed_is_true(
        self, mocked_get_boto3_client
    ):
        """When adding a route to a non-deployed gateway no post-save logic should run even is deployed is True"""
        ApiGatewayAdditionalRoute.objects.create(
            api_gateway=self.non_deployed_api_gateway,
            name="Alternative Route",
            route_key="another-route",
            integration_url="https://www.example.com/ws/other",
            deployed=True,
        )
        self.assertEqual(0, mocked_get_boto3_client.call_count)

    @patch("django_aws_api_gateway_websockets.models.get_boto3_client")
    def test_adding_route_to_gateway_with_deployment_id_but_deployed_is_true(
        self, mocked_get_boto3_client
    ):
        """When saving an additional route where the deployed property is TRUE, the create_route method should not
        be called even as we expect the route has already been deployed another way
        """
        ApiGatewayAdditionalRoute.objects.create(
            api_gateway=self.deployed_api_gateway,
            name="Alternative Route",
            route_key="another-route",
            integration_url="https://www.example.com/ws/other",
            deployed=True,
        )
        self.assertEqual(0, mocked_get_boto3_client.call_count)

    @patch("django_aws_api_gateway_websockets.models.get_boto3_client")
    def test_adding_route_to_gateway_with_deployment_id_and_deployed_is_false(
        self, mocked_get_boto3_client
    ):
        """When saving an additional route where the deployed property is False and the API Gateway has a deployment ID
        then the new additional route should be automatically deployed
        """
        mocked_get_boto3_client.return_value.create_integration.return_value = {
            "IntegrationId": "Int-ID-123"
        }

        obj = ApiGatewayAdditionalRoute.objects.create(
            api_gateway=self.deployed_api_gateway,
            name="Alternative Route",
            route_key="another-route",
            integration_url="https://www.example.com/ws/other",
            deployed=False,
        )

        self.assertTrue(obj.deployed)

        mocked_get_boto3_client.assert_called_with()
        mocked_get_boto3_client.return_value.create_integration.assert_called_with(
            ApiId=self.deployed_api_gateway.api_id,
            ConnectionType="INTERNET",
            IntegrationMethod="POST",
            IntegrationType="HTTP_PROXY",
            IntegrationUri=obj.integration_url,
            PassthroughBehavior="WHEN_NO_MATCH",
            PayloadFormatVersion="1.0",
            RequestParameters={
                "integration.request.header.connectionId": "context.connectionId"
            },
            TimeoutInMillis=29000,
        )
        mocked_get_boto3_client.return_value.create_route(
            ApiId=obj.api_gateway.api_id,
            ApiKeyRequired=False,
            AuthorizationType="NONE",
            RouteKey=obj.route_key,
            Target="integrations/Int-ID-123",
            RouteResponseSelectionExpression="$default",
        )
        mocked_get_boto3_client.return_value.create_deployment(
            ApiId=self.deployed_api_gateway.api_id,
            Description=self.deployed_api_gateway.stage_description,
            StageName=self.deployed_api_gateway.stage_name,
        )


class WebSocketSessionSimpleTestCase(SimpleTestCase):
    def test__str__returns_name(self):
        """The __str__ of WebSocketSession should return the connection_id attribute"""
        obj = WebSocketSession(connection_id="ABC-123")
        self.assertEqual("ABC-123", str(obj))


class WebSocketSessionIntegrationTestCase(TestCase):
    def setUp(self) -> None:
        self.api_gateway = ApiGateway.objects.create(
            api_name="First Gateway",
            api_description="A test api gateway",
            target_base_endpoint="http://www.example1.com/",
            deployment_id="",
        )

    @patch("django_aws_api_gateway_websockets.models.get_boto3_client")
    @override_settings(AWS_REGION_NAME="eu-west-1")
    def test_send_message_return_response_on_success(self, mocked_get_boto3_client):
        """The return of post_to_connection should be returned on a positive send_message"""
        obj = WebSocketSession.objects.create(
            connection_id="ABC-123",
            channel_name="My-Channel",
            connected=True,
            api_gateway=self.api_gateway,
        )
        data = {"action": "example", "msg": "Test Me"}

        res = obj.send_message(data)

        mocked_get_boto3_client.assert_called_with(
            "apigatewaymanagementapi",
            endpoint_url=(
                f"https://{obj.api_gateway.api_id}.execute-api."
                f"{settings.AWS_REGION_NAME}.amazonaws.com/{obj.api_gateway.stage_name}"
            ),
        )
        mocked_get_boto3_client.return_value.post_to_connection.assert_called_with(
            Data=json.dumps(data), ConnectionId=obj.connection_id
        )
        self.assertEqual(
            mocked_get_boto3_client.return_value.post_to_connection.return_value, res
        )

    @patch("django_aws_api_gateway_websockets.models.get_boto3_client")
    @override_settings(AWS_REGION_NAME="eu-west-1")
    def test_send_message_boto3_gone_away(self, mocked_get_boto3_client):
        """When a GoneException is raised by Boto3 then an exception is not bubbled but the connected property is set
        to False
        """
        mocked_get_boto3_client.return_value.post_to_connection.side_effect = (
            ClientError(
                {"Error": {"Code": "GoneException", "Message": "Actual Error Here"}},
                "apigateway",
            )
        )

        obj = WebSocketSession.objects.create(
            connection_id="ABC-123",
            channel_name="My-Channel",
            connected=True,
            api_gateway=self.api_gateway,
        )
        data = {"action": "example", "msg": "Test Me"}

        obj.send_message(data)

        mocked_get_boto3_client.assert_called_with(
            "apigatewaymanagementapi",
            endpoint_url=(
                f"https://{obj.api_gateway.api_id}.execute-api."
                f"{settings.AWS_REGION_NAME}.amazonaws.com/{obj.api_gateway.stage_name}"
            ),
        )
        mocked_get_boto3_client.return_value.post_to_connection.assert_called_with(
            Data=json.dumps(data), ConnectionId=obj.connection_id
        )

    @patch("django_aws_api_gateway_websockets.models.get_boto3_client")
    @override_settings(AWS_REGION_NAME="eu-west-1")
    def test_send_message_raises_other_boto3_exceptions(self, mocked_get_boto3_client):
        """When a GoneException is raised by Boto3 then an exception is not bubbled but the connected property is set
        to False
        """
        mocked_get_boto3_client.return_value.post_to_connection.side_effect = (
            ClientError(
                {"Error": {"Code": "ABC123", "Message": "Actual Error Here"}},
                "apigateway",
            )
        )

        obj = WebSocketSession.objects.create(
            connection_id="ABC-123",
            channel_name="My-Channel",
            connected=True,
            api_gateway=self.api_gateway,
        )
        data = {"action": "example", "msg": "Test Me"}

        with self.assertRaises(ClientError):
            obj.send_message(data)

        mocked_get_boto3_client.assert_called_with(
            "apigatewaymanagementapi",
            endpoint_url=(
                f"https://{obj.api_gateway.api_id}.execute-api."
                f"{settings.AWS_REGION_NAME}.amazonaws.com/{obj.api_gateway.stage_name}"
            ),
        )


class WebSocketSessionQuerySetIntegrationTestCase(TestCase):
    # todo - Implement these tests
    pass
